################################################################################
#   Gene prediction pipeline 
#
#   $Id: Exons.py 2881 2010-04-07 08:45:38Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
Exons.py - A library to read/write/manage exons.
=====================================================

:Author: 
:Release: $Id$
:Date: |today|
:Tags: Python

"""

import re, sys, string
try: import alignlib
except ImportError: pass

class Exon:
    """class for exons.

    contains info about the genomic location of an exon
    and its location within a peptide sequence.

    The field mAlignment is set optionally.
    """
    def __init__(self ):
        self.mQueryToken = ""
        self.mSbjctToken = ""
        self.mSbjctStrand = ""
        self.frame = 0
        self.mRank = 0
        self.mPeptideFrom = 0
        self.mPeptideTo = 0
        self.mGenomeFrom = 0
        self.mGenomeTo = 0
        self.mAlignment = "U"
        
    def Read( self, line, contig_sizes = {}, format = "exons", extract_id = None, converter = None ):
        """read exon from tab-separated line.

        extract_id is a regular expression object to extract the identifier from
        the identifier column.

        if converter is given, it is used to convert to
        zero-based open-closed both strand coordinates.
        """

        if format == "exons":
            (self.mQueryToken,
             self.mSbjctToken,
             self.mSbjctStrand,
             self.frame,
             self.mRank,
             self.mPeptideFrom,
             self.mPeptideTo,
             self.mGenomeFrom,
             self.mGenomeTo) = line[:-1].split("\t")
        elif format == "gtf":
            (self.mSbjctToken,
             a, b,
             self.mGenomeFrom,
             self.mGenomeTo,
             c, 
             self.mSbjctStrand,
             self.frame,
             self.mQueryToken ) = line[:-1].split("\t")
            
            self.mRank = 0
            self.mPeptideFrom, self.mPeptideTo = 0, 0
            self.mGenomeFrom = int(self.mGenomeFrom) - 1

        if extract_id:
            self.mQueryToken = extract_id.search( self.mQueryToken).groups()[0]
        
        if self.frame == ".": self.frame = 0
            
        (self.mRank, self.frame,
         self.mPeptideFrom, self.mPeptideTo,
         self.mGenomeFrom, self.mGenomeTo ) = map( int,\
                                                   (self.mRank, self.frame,
                                                    self.mPeptideFrom, self.mPeptideTo,
                                                    self.mGenomeFrom, self.mGenomeTo ))

        if self.mSbjctStrand == "1": self.mSbjctStrand = "+"
        elif self.mSbjctStrand in ("0", "-1"): self.mSbjctStrand = "-"

        if converter and self.mSbjctToken in contig_sizes:
            self.mGenomeFrom, self.mGenomeTo = converter( self.mGenomeFrom,
                                                          self.mGenomeTo,
                                                          self.mSbjctStrand == "+",
                                                          contig_sizes[self.mSbjctToken] )
        
        if self.mSbjctStrand == "-" and self.mSbjctToken in contig_sizes:
            sbjct_length = contig_sizes[self.mSbjctToken]                    
            self.mGenomeFrom, self.mGenomeTo = sbjct_length - self.mGenomeTo, sbjct_length - self.mGenomeFrom

        if self.mGenomeFrom > self.mGenomeTo:
            self.mGenomeFrom, self.mGenomeTo = self.mGenomeTo, self.mGenomeFrom

        return 1

    def __str__( self ):
        return string.join( map(str, (
            self.mQueryToken,
            self.mSbjctToken,
            self.mSbjctStrand,
            self.frame,
            self.mRank,
            self.mPeptideFrom,
            self.mPeptideTo,
            self.mGenomeFrom,
            self.mGenomeTo)), "\t" )

    def Merge( self, other ):
        """Merge this exon with another (adjacent and preceeding) exon.

        Do not merge if the distance between exons is not divisible by 3.
        Merging of two exons invalidated peptide coordinates for all following
        exons. These need to be updated.
        """
        
        if other.mSbjctStrand != self.mSbjctStrand or \
           other.mSbjctToken != self.mSbjctToken:
            raise ValueError, "exons not on the same contig and strand."

        if other.mGenomeTo >= self.mGenomeFrom:
            raise ValueError, "other exon not preceding this exon"

        difference_sbjct = self.mGenomeFrom - other.mGenomeTo
        difference_query = 0
        
        if difference_sbjct % 3 != 0:
            raise ValueError, "can not merge exons with incompatible phases."
        
        self.mGenomeFrom = other.mGenomeFrom
        self.mPeptideFrom = other.mPeptideFrom
        self.mPeptideTo += difference_sbjct / 3
        self.frame = other.frame
        
        ## remove split codons and add to difference
        ## introduce a gap in the peptide sequence as well
        if other.mAlignment[-1][0] == "S":
            del other.mAlignment[-1]
            del self.mAlignment[0]
            difference_sbjct += 3
            difference_query += 3

        extra = [("G", 0, difference_sbjct)]
        if difference_query:
            extra.append( ("G", difference_query, 0) )

        self.mAlignment = other.mAlignment + extra + self.mAlignment

    def InvertGenomicCoordinates( self, lgenome):
        """invert genomic alignment on sequence.

        Negative strand is calculated from the other end.
        """
        if self.mSbjctStrand == "-":
            x = min(self.mGenomeFrom, self.mGenomeTo)
            y = max(self.mGenomeFrom, self.mGenomeTo)
                
            self.mGenomeFrom = lgenome - y
            self.mGenomeTo   = lgenome - x
            
    def GetCopy( self ):
        
        e = Exon()

        (e.mQueryToken,
         e.mSbjctToken,
         e.mSbjctStrand,
         e.frame,
         e.mRank,
         e.mPeptideFrom,
         e.mPeptideTo,
         e.mGenomeFrom,
         e.mGenomeTo) = \
         (self.mQueryToken,
          self.mSbjctToken,
          self.mSbjctStrand,
          self.frame,
          self.mRank,
          self.mPeptideFrom,
          self.mPeptideTo,
          self.mGenomeFrom,
          self.mGenomeTo)
        return e

##----------------------------------------------------------
def UpdatePeptideCoordinates( exons ):
    """updates peptides coordinates for a list of exons.

    Exons have to  be sorted.
    """
    last_e = exons[0]
    
    for e in exons[1:]:
        if e.mPeptideFrom != last_e.mPeptideTo:
            offset = last_e.mPeptideTo - e.mPeptideFrom
            e.mPeptideFrom += offset
            e.mPeptideTo += offset
        last_e = e
        
##----------------------------------------------------------
def PostProcessExons( all_exons,
                      do_invert = None,
                      remove_utr = None,
                      filter = None,
                      reset = False,
                      require_increase = False,
                      no_invert = False,
                      contig_sizes = {},
                      from_zero = False,
                      delete_missing = False,
                      set_peptide_coordinates = False,
                      set_rank = False):
    """do post-processing of exons

    exons is a dictionary of lists of exons.

    Exons are sorted by mPeptideFrom.
    
    Operations include:
    
    -invert: sort out forward/reverse strand coordinates

    -set_peptide_coordinates: sets the peptide coordinates of
        exons.

    -set-rank: set rank of exons
    
    -remove_utr: remove any utr (needs peptide coordinates)

    -delete_missing: if set set true, exons on contigs not in contig_sizes
        will be deleted.
        
    -from_zero: exon genomic coordinates start at 0

    -reset: exon genomic coordinates start 0
    """
    
    for k in all_exons.keys():

        exons = all_exons[k]
        if delete_missing and contig_sizes and exons[0].mSbjctToken not in contig_sizes:
            del all_exons[k]
            continue
        
        invert = False

        ## ENSEMBL PATCH: exons on different strands, for example CG32491!
        # if exons[0].mGenomeFrom > exons[-1].mGenomeFrom and \
        #       exons[0].mSbjctStrand == exons[-1].mSbjctStrand:
        #    invert = True

        ## convert to forward/reverse strand coordinates if so desired
        if (invert or do_invert or from_zero) and not no_invert:

            if contig_sizes:
                if exons[0].mSbjctToken in contig_sizes:
                    l = contig_sizes[exons[0].mSbjctToken]
                elif "dummy" in contig_sizes:
                    l = contig_sizes["dummy"]
                else:
                    continue
                
            elif from_zero:
                l = 0
            else:
                max_from = max( map(lambda x: x.mGenomeFrom, exons))
                max_to = max( map(lambda x: x.mGenomeTo, exons))
                l = max(max_from, max_to)

            for exon in exons:
                exon.InvertGenomicCoordinates( l )

        ## set peptide coordinates. Exon genomic coordinates are now forward/reverse strand
        ## coordinates so sorting is straight-forward
        if set_peptide_coordinates:
            exons.sort( lambda x,y: cmp( x.mGenomeFrom, y.mGenomeFrom) )
            start = 0
            for exon in exons:
                exon.mPeptideFrom = start
                start += exon.mGenomeTo - exon.mGenomeFrom
                exon.mPeptideTo = start
        else:
            exons.sort( lambda x,y: cmp( x.mPeptideFrom, y.mPeptideFrom) )

        ## set rank
        if set_rank:
            rank = 1
            for exon in exons:
                exon.mRank = rank
                rank += 1

        if remove_utr:
            # remove 5' UTR
            exon = exons[0]
            exon.mGenomeFrom = exon.mGenomeTo - (exon.mPeptideTo - exon.mPeptideFrom)
            # remove 3' UTR
            exon = exons[-1]
            exon.mGenomeTo = exon.mGenomeFrom + (exon.mPeptideTo - exon.mPeptideFrom)

        ## patch for last exon
        ## Occasionally I found an incomplete codon
        ## at the end of a gene. The peptide sequence
        ## was ok (apart from including the stop codon?)
        ## Example: ENSGALP00000006767
        ## NB: happened only in chicken, why?
        ## Check if last codon is complete, otherwise
        ## add appropriate residues to genome_to
        ## dangerous: might be due to pseudogene!
##         e = exons[-1]
##         if e.mPeptideTo % 3 != 0:
##             d = 3 - e.mPeptideTo % 3
##             e.mPeptideTo += d
##             e.mGenomeTo += d

        if reset:
            offset = exons[0].mGenomeFrom
            for e in exons:
                e.mGenomeFrom -= offset
                e.mGenomeTo -= offset
                
    return all_exons

##----------------------------------------------------------
def GetExonBoundariesFromTable( dbhandle,
                                table_name_predictions = "predictions",
                                table_name_exons = "exons",
                                only_good = False,
                                do_invert = None,
                                remove_utr = None,
                                filter = None,
                                reset = False,
                                require_increase = False,
                                contig_sizes = {},
                                prediction_ids = None,
                                table_name_quality = "quality",
                                table_name_redundant = "redundant",
                                non_redundant_filter = False,
                                schema = None,
                                quality_filter = None,
                                from_zero = False,
                                delete_missing = False ):
    """get exon boundaries from table."""

    extra = ""
    extra_tables = ""
    
    if only_good:
        extra += " AND e.is_ok = TRUE"

    if schema:
        table_name_predictions = "%s.%s" % (schema, table_name_predictions)
        table_name_exons = "%s.%s" % (schema, table_name_exons)                
        table_name_redundant = "%s.%s" % (schema, table_name_redundant)
        table_name_quality = "%s.%s" % (schema,table_name_quality)
        
    if prediction_ids:
        extra += " AND p.prediction_id IN ('%s')" % ("','".join(map(str,prediction_ids)))

    if quality_filter:
        extra_tables += ", %s AS q" % table_name_quality
        extra += " AND q.prediction_id = p.prediction_id AND q.class IN ('%s')" % ("','".join(quality_filter))

    if non_redundant_filter:
        extra_tables += ", %s AS r" % table_name_redundant
        extra += " AND r.rep_prediction_id = r.mem_prediction_id AND r.rep_prediction_id = p.prediction_id"
                                                  
    statement = """
    SELECT DISTINCT p.prediction_id, p.sbjct_token, p.sbjct_strand,
    e.exon_frame, 0, e.exon_from, e.exon_to, e.genome_exon_from, e.genome_exon_to
    FROM %s AS p, %s AS e %s
    WHERE p.prediction_id = e.prediction_id AND e.exon_to > 0
    %s
    ORDER BY p.prediction_id, e.genome_exon_from
    """ % (table_name_predictions, table_name_exons, extra_tables, extra )
    cc = dbhandle.cursor()
    cc.execute(statement)
    result = cc.fetchall()
    cc.close()
    
    all_exons = {}
    
    for r in result:
        e = Exon()
        e.Read( "\t".join( map(str, r)) + "\n", contig_sizes)

        if filter and not e.mQueryToken in filter: continue

        if not all_exons.has_key( e.mQueryToken ): all_exons[e.mQueryToken] = []
        all_exons[e.mQueryToken].append( e )
        
    return PostProcessExons( all_exons,
                             do_invert,
                             remove_utr,
                             filter,
                             reset,
                             require_increase,
                             contig_sizes,
                             from_zero = from_zero,
                             delete_missing = delete_missing )


##----------------------------------------------------------
def CountNumExons( exons ):
    """return hash with number of exons per entry."""

    nexonst = {}
    for k, ee in exons.items():
        nexons[k] = len(ee)
    return nexons
        
##----------------------------------------------------------
def SetRankToPositionFlag( exons ):
    """set rank for all exons.

    Set rank to
    1 : if it is first exon,
    -1: if it is the last exon (single exon genes are -1)
    0 : if it is an internal exon.
    """

    for k, ee in exons.items():
        if len(ee) == 1:
            ee[0].mRank = -1
            continue
        ee.sort( lambda x,y: cmp( x.mPeptideFrom, y.mPeptideFrom ) )
        ee[0].mRank = 1
        ee[-1].mRank = -1
        for x in range(1, len(ee) - 1):
            ee[x].mRank = 0

##----------------------------------------------------------
def ReadExonBoundaries( file,
                        do_invert = None,
                        remove_utr = None,
                        filter = None,
                        reset = False,
                        require_increase = False,
                        no_invert = False,
                        contig_sizes = {},
                        converter = None,
                        from_zero = False,
                        delete_missing = False,
                        format = "exons",
                        gtf_extract_id = None ):
    """read exons boundaries from tab separated file.

    if remove_utr is set, the UTR of the first/last exon is removed.

    if reset is set, then the genomic part is moved so that it starts at 1.
    if require_increase is set, then exons are sorted in increasing order.

    If do_invert is set: negative strand coordinates are converted to positive
    strand coordinates

    if no_invert is set: coordinates are kept as they are.

    if from_zero is set: coordinates are mapped from 0. Thus reverse
    strand coordinates will be negative.

    if delete_missing is True and sbjct-token is not in contig_sizes
    but the exon needs to be inverted: delete transcript.

    The exon file format is tab-separated and can be of the two formats:

    format="exons":
    id, contig, strand, frame, rank, peptide_from, peptide_to, genome_from, genome_to

    format = "gtf":
    contig, ignored, ignored, genome_from, genome_to, ignored, strand, frame, id

    if converter is given, use it to convert to forward/reverse strand coordinates.

    gtg_extract_id: regular expression object to extract id from id column.
    """

    all_exons = {}

    l = None
    for line in file:
        if line[0] == "#": continue
        if line[0] == "": continue
        e = Exon()

        e.Read(line,
               contig_sizes = contig_sizes,
               format = format,
               converter = converter,
               extract_id = gtf_extract_id )
        
        l = e
        if filter and not e.mQueryToken in filter: continue
        if not all_exons.has_key( e.mQueryToken ): all_exons[e.mQueryToken] = []
        all_exons[e.mQueryToken].append( e )

    set_rank = False
    set_peptide_coordinates = False
    if format == "gtf":
        set_rank = True
        set_peptide_coordinates = True

    return PostProcessExons( all_exons,
                             do_invert = do_invert,
                             remove_utr = remove_utr,
                             filter = filter,
                             reset = reset,
                             require_increase = require_increase,
                             no_invert = no_invert,
                             contig_sizes = contig_sizes,
                             from_zero = from_zero,
                             delete_missing = delete_missing,
                             set_peptide_coordinates = set_peptide_coordinates,
                             set_rank = set_rank )

    
##------------------------------------------------------------
def Alignment2Exons( alignment, query_from = 0, sbjct_from = 0, add_stop_codon = 1):
    """convert a Peptide2DNA alignment to exon boundaries.
    """

    exons = []

    ## count in nucleotides for query
    query_from *= 3
    query_pos = query_from
    sbjct_pos = sbjct_from
    frame = 0
    ali = []
    
    for state, l_query, l_sbjct in alignment:

        if state in ("S", "M", "G", "F"):
            ali.append( (state, l_query, l_sbjct))

        ## count as nucleotides
        l_query *= 3
                
        if state == "M":
            query_pos += l_query

        elif state == "G":
            query_pos += l_query
            
        elif state == "P":
            ## state P is counted as an intron
            frame = query_from % 3
            if frame != 0: frame = 3 - frame
            exon = Exon()
            exon.mPeptideFrom = query_from
            exon.mPeptideTo = query_pos
            exon.mSbjctFrame = frame
            exon.mGenomeFrom = sbjct_from
            exon.mGenomeTo = sbjct_pos
            exon.mAlignment = ali
            exons.append( exon )
            ali = []
            query_pos += l_query
            query_from = query_pos
            sbjct_from = sbjct_pos + l_sbjct

        elif state == "S":
            query_pos += l_sbjct

        elif state == "I":
            pass
        
        elif state == "5":
            frame = query_from % 3
            if frame != 0: frame = 3 - frame
            exon = Exon()
            exon.mPeptideFrom = query_from
            exon.mPeptideTo = query_pos
            exon.frame = frame
            exon.mGenomeFrom = sbjct_from
            exon.mGenomeTo = sbjct_pos
            exon.mAlignment = ali
            exons.append( exon )
            ali = []
            query_pos += l_query
            
        elif state == "3":
            query_pos += l_query
            query_from = query_pos
            sbjct_from = sbjct_pos + l_sbjct
            
        sbjct_pos += l_sbjct
                
    ## add three for the stop codon:
    if add_stop_codon:
        query_pos += 3
        sbjct_pos += 3
        
    frame = query_from % 3
    if frame != 0: frame = 3 - frame

    exon = Exon()
    exon.mPeptideFrom = query_from
    exon.mPeptideTo = query_pos
    exon.frame = frame
    exon.mGenomeFrom = sbjct_from
    exon.mGenomeTo = sbjct_pos
    exon.mAlignment = ali
    
    exons.append( exon )

    return exons

##-------------------------------------------------------------------
def Exons2Alignment( exons ):
    """build alignment string from a (sorted) list of exons.
    """
    alignment = []
    last_e = exons[0]

    for e in exons[1:]:

        alignment += last_e.mAlignment

        difference = e.mGenomeFrom - last_e.mGenomeTo
        
        alignment.append( ("5", 0, 2) )
        alignment.append( ("I", 0, difference - 4 ) )
        alignment.append( ("3", 0, 2) )
        last_e = e

    alignment += last_e.mAlignment 
    return alignment

##------------------------------------------------------------
class ComparisonResult:

    class Link:
        def __init__(self):
            self.mId1 = 0
            self.mId2 = 0
            self.mPercentIdentity = 0
            self.mPercentSimilarity = 0
            self.mCoverage = 0
            self.mOverlap = 0
            self.mIsSameFrame = True
            self.mIsGoodExon = True

        def __str__( self ):

            return ";".join( map(str, (self.mId1, self.mId2, self.mPercentIdentity, self.mPercentSimilarity,
                                       self.mCoverage, self.mOverlap,
                                       self.mIsSameFrame,
                                       self.mIsGoodExon )))
                                 
    def __init__(self):

        self.mNumDifferenceExons = 0
        self.mSumBoundaryDifferences = 0
        self.mMaxBoundaryDifferences = 0
        self.mNumIdenticalExons  = 0
        self.mNumMissedCmpBoundaries = 0
        self.mNumMissedRefBoundaries = 0
        self.mNumCmpBoundaries = 0
        self.mNumRefBoundaries = 0         
        self.mNumDubiousExons    = 0
        self.mNumDeletedExons    = 0
        self.mNumInsertedExons   = 0
        self.mNumDeletedIntrons  = 0
        self.mNumInsertedIntrons = 0
        self.mNumTruncatedNExons = 0
        self.mNumTruncatedCExons = 0
        self.mNumExtendedNExons = 0
        self.mNumExtendedCExons = 0
        self.mNumDeletedNExons   = 0
        self.mNumDeletedCExons   = 0
        self.mNumInsertedNExons  = 0
        self.mNumInsertedCExons  = 0
        self.mNumSkippedExons    = 0
        self.mEquivalences = []

    def __str__(self):
        a = string.join( map(str, ( \
            self.mNumDifferenceExons,
            self.mNumIdenticalExons,
            self.mNumSkippedExons,
            self.mNumMissedRefBoundaries,                        
            self.mNumMissedCmpBoundaries,            
            self.mNumCmpBoundaries,
            self.mNumRefBoundaries,            
            self.mSumBoundaryDifferences,
            self.mMaxBoundaryDifferences,
            self.mNumDubiousExons    ,
            self.mNumDeletedExons    ,
            self.mNumInsertedExons   ,
            self.mNumDeletedIntrons  ,
            self.mNumInsertedIntrons ,
            self.mNumTruncatedNExons ,
            self.mNumTruncatedCExons ,
            self.mNumExtendedNExons ,
            self.mNumExtendedCExons ,
            self.mNumDeletedNExons   ,
            self.mNumDeletedCExons   ,
            self.mNumInsertedNExons  ,
            self.mNumInsertedCExons )), "\t")
        
        b = " ".join(map(str, self.mEquivalences))
        
        return a + "\t" + str(b)

    def Pretty( self, prefix = "# " ):
        lines = []
        for k,v in self.__dict__.items():
            if k[0] == 'm':
                if k == "mEquivalences":
                    for vv in v:
                        lines.append( "%s%-44s: %s" % (prefix, k, str(vv)) )
                else:
                    lines.append( "%s%-40s: %s" % (prefix, k, str(v)) )
        return string.join(lines, "\n")

    def GetHeader( self ):
        """return header line for tab-separated column format."""
        return string.join( (
            "NumDifferenceExons",
            "NumIdenticalExons",
            "NumSkippedExons",
            "NumMissedRefBoundaries",                        
            "NumMissedCmpBoundaries",            
            "NumCmpBoundaries",
            "NumRefBoundaries",            
            "SumBoundaryDifferences",
            "MaxBoundaryDifferences",
            "NumDubiousExons",
            "NumDeletedExons",
            "NumInsertedExons",
            "NumDeletedIntrons",
            "NumInsertedIntrons",
            "NumTruncatedNExons",
            "NumTruncatedCExons",
            "NumExtendedNExons",
            "NumExtendedCExons",
            "NumDeletedNExons",
            "NumDeletedCExons",
            "NumInsertedNExons",
            "NumInsertedCExons"), "\t")
        
def RemoveRedundantEntries( l ):
    """remove redundant entries (and 0s) from list.

    One liner?
    """

    if len(l) == 0:
        return l
    
    l.sort()
    last = l[0]
    n = [last]
    for x in l[1:]:
        if x != last and x > 0:
            n.append( x )
        last = x
    return n

def CompareGeneStructures( xcmp_exons, ref_exons, 
                           map_ref2cmp = None,
                           cmp_sequence = None,
                           ref_sequence = None,
                           threshold_min_pide = 0,
                           threshold_slipping_exon_boundary = 9,
                           map_cmp2ref = None,
                           threshold_terminal_exon = 15):
    
    """Compare two gene structures.

    This function is useful for comparing the exon boundaries of
    a predicted peptide with the exon boundaries of the query peptide.

    cmp_exons are exons for the gene to test.
    ref_exons are exons from the reference.
    
    Exon boundaries are already mapped to the peptide for the
    reference.
    
    map_ref2cmp: Alignment of protein sequences for cmp and ref.
    map_cmp2ref: Alignment of cmp to ref. If given, mapping is done from cmp to ref.
    Invalid exon boundaries can be set to -1.

    threshold_terminal_exon:
        Disregard terminal exons for counting missed boundaries, if they are
        maximum x nucleotides long.
    
    
    """

    if ref_sequence and cmp_sequence:
        cmp_seq = alignlib.makeSequence( cmp_sequence)
        ref_seq = alignlib.makeSequence( ref_sequence)
    else:
        ref_seq, cmp_seq = None, None
        
    result = ComparisonResult()
    # unmappable exons have to be kept for counting purpses.
    if map_cmp2ref:
        cmp_exons = MapExons( xcmp_exons, map_cmp2ref )
    else:
        cmp_exons = xcmp_exons

    cmp_len = cmp_exons[-1].mPeptideTo
    ref_len = ref_exons[-1].mPeptideTo    
    
    # default values
    result.mNumDifferenceExons = len(ref_exons) - len(cmp_exons)

    # number of matching boundaries (do not use first and last)
    boundaries_cmp = map( lambda x: x.mPeptideFrom, cmp_exons) + map( lambda x: x.mPeptideTo, cmp_exons)
    boundaries_ref = map( lambda x: x.mPeptideFrom, ref_exons) + map( lambda x: x.mPeptideTo, ref_exons)
    
    boundaries_cmp = RemoveRedundantEntries( boundaries_cmp[1:-1] )
    boundaries_ref = RemoveRedundantEntries( boundaries_ref[1:-1] )    
    
    result.mNumMissedCmpBoundaries = CountMissedBoundaries( boundaries_cmp, boundaries_ref,
                                                            threshold_slipping_exon_boundary,
                                                            threshold_terminal_exon, cmp_len - threshold_terminal_exon + 1 )
    result.mNumMissedRefBoundaries = CountMissedBoundaries( boundaries_ref, boundaries_cmp,
                                                            threshold_slipping_exon_boundary,
                                                            threshold_terminal_exon, ref_len - threshold_terminal_exon + 1 )                                                            
    result.mNumCmpBoundaries = len(boundaries_cmp)    
    result.mNumRefBoundaries = len(boundaries_ref)

    in_sync = False

    # assume that peptides start at 1.
    cmp_from = 1
    ref_from = 1
    cmp_to = 0
    for e in cmp_exons:
        cmp_to = max( cmp_to, e.mPeptideTo)
    cmp_to = cmp_to / 3 + 1
    ref_to = 0
    for e in ref_exons:
        ref_to = max( ref_to, e.mPeptideTo)
    ref_to = ref_to / 3 + 1

    e,r = 0,0

    while e < len(cmp_exons) and r < len(ref_exons):
        
        percent_identity = 0
        percent_similarity = 0
        is_good_exon = False

        exon1 = cmp_exons[e]
        exon2 = ref_exons[r]
        
        if e < len(cmp_exons)-1 :
            next_cmp_exon = cmp_exons[e+1]
        else:
            next_cmp_exon = None

        if r < len(ref_exons) - 1:
            next_ref_exon = ref_exons[r+1]
        else:
            next_ref_exon = None

        ## cmp can have 0,0 entries, these are exons that were impossible to align
        if exon1.mPeptideFrom == 0 and exon1.mPeptideTo == 0:
            if r == 0:
                result.mNumInsertedNExons += 1
            else:
                result.mNumInsertedExons += 1                
            e += 1
        elif exon1.mPeptideFrom < 0 or exon1.mPeptideTo < 0:
            result.mNumSkippedExons += 1
            e += 1
        elif exon2.mPeptideFrom < 0 or exon2.mPeptideTo < 0:
            result.mNumSkippedExons += 1
            r += 1
        elif exon2.mPeptideTo <= exon1.mPeptideFrom + threshold_slipping_exon_boundary:
            ## no overlap
            if e == 0:
                result.mNumDeletedNExons += 1
            else:
                result.mNumDeletedExons += 1
            r += 1
        elif exon1.mPeptideTo <= exon2.mPeptideFrom + threshold_slipping_exon_boundary:
            ## no overlap
            if r == 0:
                result.mNumInsertedNExons += 1
            else:
                result.mNumInsertedExons += 1
            e += 1
        else:
            ## overlap
            dfrom = int(abs(exon1.mPeptideFrom - exon2.mPeptideFrom))
            dto = int(abs(exon1.mPeptideTo - exon2.mPeptideTo))

            overlap  = min(exon1.mPeptideTo,  exon2.mPeptideTo) - max(exon1.mPeptideFrom, exon2.mPeptideFrom)
            coverage = 100 * overlap / (max(exon1.mPeptideTo,  exon2.mPeptideTo) - min(exon1.mPeptideFrom, exon2.mPeptideFrom))
            
            is_good_exon = True

            link = ComparisonResult.Link()
            link.mId1 = e
            link.mId2 = r
            link.mIsGoodExon = is_good_exon
            link.mIsSameFrame = exon1.frame = exon2.frame
            link.mCoverage = coverage
            link.mOverlap = overlap
            
            ## get percent identity
            if cmp_seq and ref_seq and map_ref2cmp:
                
                tmp_ali = alignlib.makeAlignmentVector()

                xquery_from = max( exon2.mPeptideFrom / 3, exon1.mPeptideFrom / 3) 
                xquery_to = min(1 + exon2.mPeptideTo / 3, 1 + exon1.mPeptideTo / 3)

                alignlib.copyAlignment( tmp_ali, map_ref2cmp, xquery_from, xquery_to )

                percent_identity = alignlib.calculatePercentIdentity( tmp_ali,
                                                                      ref_seq,
                                                                      cmp_seq ) * 100
                percent_similarity = alignlib.calculatePercentSimilarity( tmp_ali ) * 100

                if percent_identity <= threshold_min_pide:
                    is_good_exon = False

                link.mPercentIdentity = percent_identity
                link.mPercentSimilarity = percent_similarity

            result.mEquivalences.append( link )
            
            ## adjust regions for terminal exons
            if e == 0 and r == 0:
                if dfrom <= (cmp_from - 1) * 3 and dfrom > 0:
                    if is_good_exon:                        
                        result.mNumTruncatedNExons = dfrom
                    dfrom = 0
                elif dfrom <= (ref_from - 1) * 3 and dfrom > 0:                
                    if is_good_exon:                        
                        result.mNumExtendedNExons = dfrom
                    dfrom = 0
                
            if e == len(cmp_exons)-1 and r == len(ref_exons)-1:
                if dto <= (ref_to - cmp_to) * 3 and dto > 0:
                    if is_good_exon:
                        result.mNumTruncatedCExons = dto
                    dto = 0
                elif dto <= (cmp_to - ref_to) * 3 and dto > 0:
                    if is_good_exon:
                        result.mNumExtendedCExons = dto
                    dto = 0

            ## do not count deviations for terminal query exons
            if e == 0 and dfrom <= (cmp_from - 1) * 3 and dfrom > 0:
                dfrom = 0

            if e == len(cmp_exons)-1 and dto <= (ref_to - cmp_to) * 3 and dto > 0:
                dto = 0
                
            ## deal with different boundary conditions:
            if dfrom <= threshold_slipping_exon_boundary and dto <= threshold_slipping_exon_boundary:
                if is_good_exon: result.mNumIdenticalExons += 1
                e += 1
                r += 1
            ## next exon within this exon2_exon
            elif exon1.mPeptideTo < exon2.mPeptideTo and \
                     next_cmp_exon and \
                     next_cmp_exon.mPeptideTo <= exon2.mPeptideTo + threshold_slipping_exon_boundary:
                if is_good_exon: result.mNumInsertedIntrons += 1
                e += 1
                in_sync = True
                dto = 0
            ## next exon2_exon within this exon
            elif exon2.mPeptideTo < exon1.mPeptideTo and \
                     next_ref_exon and \
                     next_ref_exon.mPeptideTo <= exon1.mPeptideTo + threshold_slipping_exon_boundary:
                if is_good_exon: result.mNumDeletedIntrons += 1
                r += 1
                in_sync = True
                dto = 0
            else:
                e += 1
                r += 1
                if in_sync:
                    dfrom = 0

            if is_good_exon:
                result.mSumBoundaryDifferences += dfrom + dto
                result.mMaxBoundaryDifferences = max( dfrom, result.mMaxBoundaryDifferences )
                result.mMaxBoundaryDifferences = max( dto, result.mMaxBoundaryDifferences )                
            else:
                result.mNumDubiousExons += 1
                
    while e < len(cmp_exons):
        e += 1
        result.mNumInsertedCExons += 1

    while r < len(ref_exons):
        r += 1        
        result.mNumDeletedCExons += 1

    return result

###################################################################################
def MapExons( exons, map_a2b ):
    """map peptide coordinates of exons with map.

    returns a list of mapped exons.
    """

    long_map = alignlib.makeAlignmentVector()

    for x in range(map_a2b.getRowFrom(), map_a2b.getRowTo() + 1):
        y = map_a2b.mapRowToCol(x)
        if y:
            alignlib.addDiagonal2Alignment( long_map, 3 * (x - 1) + 1, 3 * x, 3 * (y - x) )

    def MyMapLeft( a, x):
        while x >= a.getRowFrom():
            c = a.mapRowToCol( x ) 
            if c: return c
            x -= 1
        else:
            return 0

    def MyMapRight( a, x):
        while x <= a.getRowTo():
            c = a.mapRowToCol( x ) 
            if c: return c
            x += 1
        else:
            return 0

    new_exons = []
    for e in exons:
        
        a = MyMapRight( long_map, e.mPeptideFrom + 1)
        b = MyMapLeft( long_map, e.mPeptideTo + 1)        
        
        ne = e.GetCopy()
        ne.mPeptideFrom, ne.mPeptideTo = a-1, b-1
        new_exons.append( ne )
            
    return new_exons

###################################################################################
def CountMissedBoundaries( cmp_boundaries, reference_boundaries,
                           max_slippage = 9,
                           min_from = 0,
                           max_to = 0):
    """count missed boundaries comparing cmp to ref.
    """

    nmissed = 0
    last_x = None
    ## check if all exon boundaries are ok
    for x in cmp_boundaries:
        if x <= min_from: continue
        if max_to and x >= max_to: continue
        is_ok = 0
        for c in reference_boundaries:
            if abs(x-c) < max_slippage:
                is_ok = 1
                break
        if not is_ok:
            nmissed += 1
                
    return nmissed

###################################################################################
def GetExonsRange( exons,
                   first, last,
                   full = True,
                   min_overlap = 0,
                   min_exon_size = 0 ):
    """get exons in range (first:last) (peptide coordinates).

    Set full to False, if you don't require full coverage.
    """

    new = []
    me = 3 * min_exon_size
    mo = 3 * min_overlap
    for e in exons:
        if e.mPeptideFrom > last or e.mPeptideTo < first:
            continue
        if full and ( e.mPeptideTo > last or e.mPeptideFrom < first):
            continue
        overlap = min(e.mPeptideTo, last) - max(e.mPeptideFrom, first)
        if overlap < mo: continue
        if e.mPeptideTo - e.mPeptideFrom < me: continue
        new.append( e.GetCopy() )
        
    return new

###################################################################################
def ClusterByExonIdentity( exons,
                           max_terminal_num_exons = 3,
                           min_terminal_exon_coverage = 0.0,
                           max_slippage = 0,
                           loglevel = 0):

    """build clusters of transcripts with identical exons.

    The boundaries in the first/last exon can vary.

    Returns two maps map_cluster2transcripts and
    map_transcript2cluster
    """

    ########################################################
    ## build maps
    num_exons = {}
    map_transcript2cluster = {}
    map_cluster2transcripts = {}
    list_of_exons = []
    
    for k, ee in exons.items():
        if k not in map_transcript2cluster:
            map_transcript2cluster[k] = k
            map_cluster2transcripts[k] = [k,]
        num_exons[k] = len(ee)
        
        for e in ee:
            list_of_exons.append( e )

    ########################################################
    ## sort exons
    list_of_exons.sort( lambda x, y: cmp( (x.mSbjctToken, x.mSbjctStrand, x.mGenomeFrom, x.mGenomeTo),
                                          (y.mSbjctToken, y.mSbjctStrand, y.mGenomeFrom, y.mGenomeTo)))

    if loglevel >= 1:
        print "# sorted %i exons" % len(list_of_exons)

    ########################################################
    ## cluster by overlap
    l = list_of_exons[0]
    last_id = l.mQueryToken
    
    for e in list_of_exons[1:]:
        
        identity = False
        coverage = 0
        overlap = 0

        ## check if they are already clustered:
        last_cluster = map_transcript2cluster[last_id]
        if e.mQueryToken in map_cluster2transcripts[last_cluster]:
            l = e
            last_id = e.mQueryToken
            continue
        
        if l.mSbjctStrand == e.mSbjctStrand and \
           l.mSbjctToken == e.mSbjctToken:
            
            ## overlap and coverage of larger exon
            overlap = min(e.mGenomeTo, l.mGenomeTo) - max( e.mGenomeFrom, l.mGenomeFrom)
            coverage = float(overlap) / max( (e.mGenomeTo-e.mGenomeFrom), (l.mGenomeTo-l.mGenomeFrom) )

            left_ok  = abs( l.mGenomeFrom - e.mGenomeFrom ) <= max_slippage
            right_ok = abs( l.mGenomeTo   - e.mGenomeTo ) <= max_slippage
            if overlap > 0:
                ## join if exon boundaries are identical
                if left_ok and right_ok:
                    identity = True
                ## join single exon genes if minimum coverage 
                elif (num_exons[e.mQueryToken] == 1 or num_exons[l.mQueryToken] == 1) and \
                         coverage >= min_terminal_exon_coverage:
                    identity = True
                ## join if one boundary is correct for terminal exons
                ## (for up to three exon genes)
                elif ( left_ok or right_ok ) and \
                      (num_exons[e.mQueryToken] < max_terminal_num_exons and \
                       num_exons[l.mQueryToken] < max_terminal_num_exons) and \
                      coverage >= min_terminal_exon_coverage :
                    identity = True
                ## join, if all exons are overlapping (minimum coverage)
                elif CheckContainedAinB( exons[l.mQueryToken], exons[e.mQueryToken],
                                         min_terminal_exon_coverage, loglevel=loglevel) or \
                     CheckContainedAinB( exons[e.mQueryToken], exons[l.mQueryToken],
                                         min_terminal_exon_coverage, loglevel=loglevel): 
                     identity = True
                    
        if loglevel >= 3:
            print "# ClusterByExonIdentity"
            print "#", identity, overlap, coverage, overlap, num_exons[e.mQueryToken], num_exons[l.mQueryToken]
            print "# l=", str(l)
            print "# e=", str(e)
            
        if identity:
            ## add current cluster to previous cluster.
            this_cluster = map_transcript2cluster[e.mQueryToken]
            map_cluster2transcripts[last_cluster] += map_cluster2transcripts[this_cluster]
            for x in map_cluster2transcripts[this_cluster]:
                map_transcript2cluster[x] = last_cluster
            map_cluster2transcripts[this_cluster] = []
            # print e.mQueryToken, last_cluster, map_cluster2transcripts

        l = e
        last_id = e.mQueryToken

    return map_cluster2transcripts, map_transcript2cluster

##----------------------------------------------------------------------------------------
def ClusterByExonOverlap( exons,
                          min_overlap = 0,
                          min_min_coverage = 0,
                          min_max_coverage = 0,
                          loglevel = 0 ):
    """build clusters of transcripts with overlapping exons.

    Exons need not be identical.

    Returns two maps map_cluster2transcripts and
    map_transcript2cluster
    """
    
    map_transcript2cluster = {}
    map_cluster2transcripts = {}
    list_of_exons = []

    for k, ee in exons.items():
        if k not in map_transcript2cluster:
            map_transcript2cluster[k] = k
            map_cluster2transcripts[k] = [k,]
        
        for e in ee:
            list_of_exons.append( e )
    
    list_of_exons.sort( lambda x, y: cmp( (x.mSbjctToken, x.mSbjctStrand, x.mGenomeFrom),
                                          (y.mSbjctToken, y.mSbjctStrand, y.mGenomeFrom)))

    if loglevel >= 1:
        print "# sorted %i exons" % len(list_of_exons)

    last_id = list_of_exons[0].mQueryToken
    last_from = list_of_exons[0].mGenomeFrom
    last_to = list_of_exons[0].mGenomeTo
    last_token = list_of_exons[0].mSbjctToken
    last_strand = list_of_exons[0].mSbjctStrand

    for e in list_of_exons[1:]:

        last_cluster = map_transcript2cluster[last_id]
        if e.mQueryToken in map_cluster2transcripts[last_cluster]:
            last_id = e.mQueryToken
            last_strand = e.mSbjctStrand
            last_token = e.mSbjctToken
            last_from = e.mGenomeFrom
            last_to = e.mGenomeTo
            continue

        o = min(e.mGenomeTo, last_to) - max(e.mGenomeFrom, last_from)
        o1 = float(o) / (e.mGenomeTo - e.mGenomeFrom)
        o2 = float(o) / (last_to - last_from)

        if last_strand == e.mSbjctStrand and \
               last_token == e.mSbjctToken and \
               o >= min_overlap and \
               min(o1,o2) >= min_min_coverage and \
               max(o1,o2) >= min_max_coverage:
            
            if loglevel >= 3:
                print "# ClusterByExonOverlap %i(%5.2f/%5.2f) between %s and %s: %i-%i with %i-%i" % \
                      ( o, o1, o2,
                        last_id, e.mQueryToken, last_from, last_to, e.mGenomeFrom, e.mGenomeTo)
                sys.stdout.flush()
            
            if e.mQueryToken not in map_cluster2transcripts[last_cluster]:
                this_cluster = map_transcript2cluster[e.mQueryToken]
                map_cluster2transcripts[last_cluster] += map_cluster2transcripts[this_cluster]
                for x in map_cluster2transcripts[this_cluster]:
                    map_transcript2cluster[x] = last_cluster
                map_cluster2transcripts[this_cluster] = []
                    
            last_from = min(e.mGenomeFrom, last_from)
            last_to = max(e.mGenomeTo, last_to)
        else:
            last_id = e.mQueryToken
            last_strand = e.mSbjctStrand
            last_token = e.mSbjctToken
            last_from = e.mGenomeFrom
            last_to = e.mGenomeTo
            
    return map_cluster2transcripts, map_transcript2cluster
    
##----------------------------------------------------------
def CheckOverlap( exons1,
                  exons2,
                  min_overlap = 1):
    """check if exons overlap.

    (does not check chromosome and strand.)
    """

    ## this could be quicker if sorted, but don't want to sort inplace.
    ## there are never that many exons anyway.
    for e1 in exons1:
        for e2 in exons2:
            if min(e1.mGenomeTo, e2.mGenomeTo) - max(e1.mGenomeFrom, e2.mGenomeFrom) >= min_overlap:
                return True
    return False

##----------------------------------------------------------
def CheckCoverage( exons1, exons2,
                   max_terminal_num_exons = 3,
                   min_terminal_exon_coverage = 0.0,
                   max_slippage = 0):
    """check if one set of exons covers the other.
    
    Note: does not check chromosome and strand, just genomic coordinates.
    """

    if len(exons1) > len(exons2):
        ee1, ee2 = exons1, exons2
    else:
        ee1, ee2 = exons2, exons1

    nexons1 = len(ee1)
    nexons2 = len(ee2)

    for e2 in ee2:
        identity = False
        
        for e1 in ee1:
            overlap = min(e1.mGenomeTo, e2.mGenomeTo) - max(e1.mGenomeFrom, e2.mGenomeFrom)

            if overlap <= 0: continue
            
            ## coverage of larger exon
            coverage = float(overlap) / max( (e1.mGenomeTo-e1.mGenomeFrom), (e2.mGenomeTo-e2.mGenomeFrom) )

            if overlap > 0:
                left_ok  = abs( e1.mGenomeFrom - e2.mGenomeFrom ) <= max_slippage
                right_ok = abs( e1.mGenomeTo   - e2.mGenomeTo ) <= max_slippage
                
                if e1.mRank == 0 and e2.mRank == 0:
                    ## for internal exons: the boundaries have to be within slippage distance
                    if left_ok and right_ok:
                        identity = True
                else:
                    ## for terminal exons and less than 4 exons, at least one
                    ## boundary has to be exact
                    if (left_ok or right_ok) and \
                        (nexons1 <= max_terminal_num_exons and nexons2 >= max_terminal_num_exons) and \
                        coverage >= min_terminal_exon_coverage :
                        identity = True

            if identity: break
            
        if not identity:
            return False

    return True

##----------------------------------------------------------
def CheckContainedAinB( exons1, exons2,
                        min_terminal_exon_coverage = 0.0,
                        loglevel = 0 ):
    """check if all exons in exons1 are contained in exons2.

    Note: does not check contig and strand.
    """
    
    for e1 in exons1:
        found = False
        for e2 in exons2:
            overlap = min(e1.mGenomeTo, e2.mGenomeTo) - max(e1.mGenomeFrom, e2.mGenomeFrom)
            ## coverage of smaller exon
            min_length = min((e1.mGenomeTo-e1.mGenomeFrom), (e2.mGenomeTo-e2.mGenomeFrom) )
            if min_length > 0:
                coverage = float(overlap) / min_length
            else:
                if overlap > 0:
                    coverage = 1
                else:
                    coverage = 0
                    
            if coverage >= min_terminal_exon_coverage:
                found = True
                break
        if not found: return False
    return True
    
##----------------------------------------------------------
def CheckCoverageAinB( exons1, exons2,
                       min_terminal_num_exons = 3,
                       min_terminal_exon_coverage = 0.0,
                       max_slippage = 0,
                       loglevel = 0 ):
    """check if exons1 are all in exons2
    
    Note: does not check contig and strand.
    """

    nexons1 = len(exons1)
    nexons2 = len(exons2)

    for e1 in exons1:
        identity = False
        
        for e2 in exons2:
            overlap = min(e1.mGenomeTo, e2.mGenomeTo) - max(e1.mGenomeFrom, e2.mGenomeFrom)
            ## coverage of larger exon
            coverage = float(overlap) / max( (e1.mGenomeTo-e1.mGenomeFrom), (e2.mGenomeTo-e2.mGenomeFrom) )

            left_ok  = abs( e1.mGenomeFrom - e2.mGenomeFrom ) <= max_slippage
            right_ok = abs( e1.mGenomeTo   - e2.mGenomeTo ) <= max_slippage

            if overlap > 0:
                if left_ok and right_ok:
                    identity = True
                    # terminal exons
                elif left_ok or right_ok:
                    if (e1.mRank != 0 or e2.mRank != 0) and \
                           coverage >= min_terminal_exon_coverage and \
                           nexons1 >= min_terminal_num_exons and \
                           nexons2 >= min_terminal_num_exons:
                        identity = True
                        
            if identity: break

        if loglevel >= 2:
            if identity:
                print "# exon found", str(e1), coverage
            else:
                print "# exon not found", str(e1), coverage
            
        if not identity:
            return False

    return True

##----------------------------------------------------------
def GetPeptideLengths( exons ):
    """for all exons get maximum length in coding nucleotides."""

    lengths = {}
    for k, ee in exons.items():
        lengths[k] = max( map( lambda x: x.mPeptideTo, ee) ) / 3 + 1
        
    return lengths
        
##----------------------------------------------------------
def GetGenomeLengths( exons ):
    """for all exons get maximum nucleotide."""

    lengths = {}
    for k, ee in exons.items():
        if ee[0].mGenomeFrom < ee[0].mGenomeTo:
            mi = min( map( lambda x: x.mGenomeFrom, ee) )
            ma = max( map( lambda x: x.mGenomeTo, ee) )        
        else:
            mi = min( map( lambda x: x.mGenomeTo, ee) )
            ma = max( map( lambda x: x.mGenomeFrom, ee) )        
            
        lengths[k] = ma - mi
        
    return lengths
        
##----------------------------------------------------------
def CalculateStats( exons ):
    """calculate some statistics for all exons.

    minimum/maximum intron/exon length, number of exons
    gene length
    """
    stats = {}
    for k, ee in exons.items():
        r = {}
        ee.sort( lambda x,y: cmp(x.mGenomeFrom, y.mGenomeFrom) )
        
        last_to = ee[0].mGenomeTo
        max_i = 0
        min_i = 0
        max_e = ee[0].mGenomeTo - ee[0].mGenomeFrom
        min_e = max_e
        for e in ee[1:]:
            max_e = max( max_e, e.mGenomeTo - e.mGenomeFrom )
            min_e = min( min_e, e.mGenomeTo - e.mGenomeFrom )
            max_i = max( max_i, e.mGenomeFrom - last_to )
            if min_i:
                min_i = min( min_i, e.mGenomeFrom - last_to )
            else:
                min_i = e.mGenomeFrom - last_to
                
        r['NumExons'] = len(ee)
        r['MaxIntronLength'] = max_i
        r['MinIntronLength'] = min_i
        r['MaxExonLength'] = max_e
        r['MinExonLength'] = min_e
        r['GeneLength'] = ee[-1].mGenomeTo - ee[0].mGenomeFrom
        
        stats[k] = r
        
    return stats
        
##----------------------------------------------------------        
def MatchExons( map_a2b, in_exons1, in_exons2, threshold_slipping_boundary = 9):
    """returns a list of overlapping exons (mapped via map_a2b)."""

    cmp_exons = MapExons( in_exons1, map_a2b )
    ref_exons = in_exons2
    
    e,r = 0,0
    matches = []
    
    while e < len(cmp_exons) and r < len(ref_exons):
        
        exon1 = cmp_exons[e]
        exon2 = ref_exons[r]

        ## no overlap
        ## cmp can have 0,0 entries, these are exons that were impossible to align
        if exon1.mPeptideFrom == 0 and exon1.mPeptideTo == 0:
            e += 1
        elif exon1.mPeptideFrom < 0 or exon1.mPeptideTo < 0:
            e += 1
        elif exon2.mPeptideFrom < 0 or exon2.mPeptideTo < 0:
            r += 1
        else:
            overlap  = min(exon1.mPeptideTo,  exon2.mPeptideTo) - max(exon1.mPeptideFrom, exon2.mPeptideFrom)
            if overlap > 0:
                matches.append( (e,r) )

            if exon1.mPeptideTo < exon2.mPeptideTo:
                e += 1
            elif exon2.mPeptideTo < exon1.mPeptideTo:
                r += 1
            else:
                e += 1
                r += 1

    return matches

###################################################################################
    



################################################################################
#   Gene prediction pipeline 
#
#   $Id: AlignedPairs.py 2781 2009-09-10 11:33:14Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
AlignedPairs.py - 
==================================================

:Author: 
:Release: $Id$
:Date: |today|
:Tags: Python

"""
import sys, os, string
try: import alignlib
except ImportError: pass

import WrapperDialign
import WrapperDBA
import WrapperBlastZ
import Genomics

# Wrapper for clustal needs to be updated
# import WrapperClustal


class Error(Exception):
    """Base class for exceptions in this module."""
    pass

    def __str__(self):
        return str(self.message)

class InputError(Error):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class AlignmentError(Error):
    """Exception raised by inconsistent alignments.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
    def _get_message(self, message): return self._message
    def _set_message(self, message): self._message = message
    message = property(_get_message, _set_message)

class UnalignedPair:

    def __init__(self, other = None, sequence1 = "", sequence2 = "", token1 = 0, token2 = 0):

        if other:
            (self.mCategory, self.mMethod,
             self.mToken1, self.mId1, self.mNum1, self.mLen1,
             self.mToken2, self.mId2, self.mNum2, self.mLen2,
             self.mSequence1, self.mSequence2,
             self.mFrom1, self.mTo1,
             self.mFrom2, self.mTo2 ) = \
             (other.mCategory, other.mMethod,
              other.mToken1, other.mId1, other.mNum1, other.mLen1,
              other.mToken2, other.mId2, other.mNum2, other.mLen2,
              other.mSequence1, other.mSequence2,
              other.mFrom1, other.mTo1,
              other.mFrom2, other.mTo2 )
        else:
            self.mCategory = ""
            self.mMethod = "unaligned"
            self.mToken1 = token1
            self.mId1 = 0
            self.mNum1 = 0
            self.mLen1 = 0
            self.mToken2 = token2
            self.mId2 = 0
            self.mNum2 = 0
            self.mLen2 = 0
            self.mSequence1 = ""
            self.mSequence2 = ""        
            self.mFrom1 = 0
            self.mTo1 = 0
            self.mFrom2 = 0
            self.mTo2 = 0

        if sequence1:
            self.mSequence1 = sequence1
            self.mFrom1 = 0
            self.mTo1 = len(sequence1)
            
        if sequence2:
            self.mSequence2 = sequence2        
            self.mFrom2 = 0
            self.mTo2 = len(sequence2)
        
    def Read( self, line ):
        """read data from tab-separated line."""
        
        data = string.split( line[:-1], "\t")
        
        (self.mCategory, self.mMethod,
         self.mToken1, self.mId1, self.mNum1, self.mLen1,
         self.mToken2, self.mId2, self.mNum2, self.mLen2,
         self.mFrom1, self.mTo1, self.mSequence1, 
         self.mFrom2, self.mTo2, self.mSequence2, ) = data[:16]

        ( self.mId1, self.mNum1, self.mLen1,
          self.mId2, self.mNum2, self.mLen2,
          self.mFrom1, self.mTo1,
          self.mFrom2, self.mTo2) = \
          map(int, ( self.mId1, self.mNum1, self.mLen1,
                     self.mId2, self.mNum2, self.mLen2,
                     self.mFrom1, self.mTo1,
                     self.mFrom2, self.mTo2, ))
        
    def __str__( self ):
        return "\t".join( map(str,\
                              ( self.mCategory, self.mMethod,
                                self.mToken1, self.mId1, self.mNum1, self.mLen1,
                                self.mToken2, self.mId2, self.mNum2, self.mLen2,
                                self.mFrom1, self.mTo1, self.mSequence1,
                                self.mFrom2, self.mTo2, self.mSequence2 ) ) )
                              

    def GetHeader( self ):
        return "\t".join( \
            ( "CATEGORY", "METHOD",
              "TOKEN1", "ID1", "TOTAL1", "LEN1",
              "TOKEN2", "ID2", "TOTAL2", "LEN2",               
              "FROM1", "TO1", "SEQ1",
              "FROM2", "TO2", "SEQ2" ) )


    def GetLegend( self ):
        return """# CATEGORY:       category
# METHOD:         alignment method
# TOKEN1:         name
# ID1:            id of segment
# TOTAL1:         number of segments
# LEN1:           length of segment
# TOKEN2:         name
# ID2:            id of segment
# TOTAL2:         number of segments
# LEN2:           length of segment
# FROM1:          start of segment1
# TO1:            end of segment1
# SEQ1:           sequence of segment1
# FROM2:          start of segment1
# TO2:            end of segment1
# SEQ2:           sequence of segment1"""

        
class AlignedPair( UnalignedPair ):

    mOptionsBlastZ = "C=2 B=0 T=0 W=6 K=2200"
    mOptionsDialign = "-n"
    mOptionsDialignLGS = "-n -it -thr 2 -lmax 30 -smin 8"

    def __init__(self, other = None, sequence1="", sequence2=""):
        
        UnalignedPair.__init__( self, other, sequence1, sequence2 )

        if other and isinstance(other, AlignedPair ):
            ( self.mIdentical,
              self.mTransitions,
              self.mTransversions,
              self.mNumGaps,
              self.mLength,
              self.mAligned,
              self.mUnaligned,
              self.mAlignmentFrom1,
              self.mAlignmentTo1,
              self.mAlignment1,
              self.mAlignmentFrom2,
              self.mAlignmentTo2,
              self.mAlignment2,
              self.mUnaligned,
              self.mBlockSizes,
              self.mGapSizes1,
              self.mGapSizes2 ) = \
              ( other.mIdentical,
                other.mTransitions,
                other.mTransversions,
                other.mNumGaps,
                other.mLength,
                other.mAligned,
                other.mUnaligned,
                other.mAlignmentFrom1,
                other.mAlignmentTo1,
                other.mAlignment1,
                other.mAlignmentFrom2,
                other.mAlignmentTo2,
                other.mAlignment2,
                other.mUnaligned,
                other.mBlockSizes,
                other.mGapSizes1,
                other.mGapSizes2 ) 
        else:
            self.mIdentical = 0
            self.mTransitions = 0
            self.mTransversions = 0
            self.mNumGaps = 0
            self.mLength = 0
            self.mAligned = 0        
            self.mUnaligned = 1
            self.mAlignmentFrom1 = 0
            self.mAlignmentTo1 = 0
            self.mAlignment1 = ""
            self.mAlignmentFrom2 = 0
            self.mAlignmentTo2 = 0
            self.mAlignment2 = ""        
            self.mBlockSizes = []
            self.mGapSizes1 = []
            self.mGapSizes2 = []


    ##-----------------------------------------------------------------------------------
    def __str__(self):
        """string representation of object."""
        return UnalignedPair.__str__(self) + "\t" +\
               "\t".join( map( str, \
                               ( self.mIdentical,
                                 self.mTransitions,
                                 self.mTransversions,
                                 self.mNumGaps,
                                 self.mLength,
                                 self.mAligned,
                                 self.mUnaligned,
                                 self.mAlignmentFrom1,
                                 self.mAlignmentTo1,
                                 self.mAlignment1,
                                 self.mAlignmentFrom2,
                                 self.mAlignmentTo2,
                                 self.mAlignment2,
                                 ",".join(map(str,self.mBlockSizes)),
                                 ",".join(map(str,self.mGapSizes1)),
                                 ",".join(map(str,self.mGapSizes2)),
                                 ) ) )

    ##-----------------------------------------------------------------------------------
    def Read( self, line ):
        """read data from tab-separated line."""
        UnalignedPair.Read( self, line )
        
        data = string.split( line[:-1], "\t")

        ( self.mIdentical,
          self.mTransitions,
          self.mTransversions,
          self.mNumGaps,
          self.mLength,
          self.mAligned,
          self.mUnaligned,
          self.mAlignmentFrom1,
          self.mAlignmentTo1,
          self.mAlignment1,
          self.mAlignmentFrom2,
          self.mAlignmentTo2,
          self.mAlignment2,
          self.mBlockSizes,
          self.mGapSizes1,
          self.mGapSizes2) = data[16:]

        ( self.mIdentical,
          self.mTransitions,
          self.mTransversions,
          self.mNumGaps,
          self.mLength,
          self.mAligned,
          self.mUnaligned,
          self.mAlignmentFrom1,
          self.mAlignmentTo1,
          self.mAlignmentFrom2,
          self.mAlignmentTo2 ) = map(int, \
                                     ( self.mIdentical,
                                       self.mTransitions,
                                       self.mTransversions,
                                       self.mNumGaps,
                                       self.mLength,
                                       self.mAligned,
                                       self.mUnaligned,
                                       self.mAlignmentFrom1,
                                       self.mAlignmentTo1,
                                       self.mAlignmentFrom2,
                                       self.mAlignmentTo2 ))

        self.mBlockSizes = map(int, self.mBlockSizes.split(","))
        if self.mGapSizes1:
            self.mGapSizes1 = map(int, self.mGapSizes1.split(","))
        else:
            self.mGapSizes1 = []
        if self.mGapSizes2:
            self.mGapSizes2 = map(int, self.mGapSizes2.split(","))
        else:
            self.mGapSizes2 = []
        
    ##-----------------------------------------------------------------------------------
    def GetHeader( self ):
        return UnalignedPair.GetHeader(self) + "\t" + \
               "\t".join( ( \
            "IDENTICAL",
            "TRANSITIONS",
            "TRANSVERSIONS",
            "NUMGAPS",
            "LENGTH",
            "ALIGNED",
            "UNALIGNED",
            "ALIFROM1", "ALITO1", "ALIGNED1",
            "ALIFROM2", "ALITO2", "ALIGNED2",
            "BLOCKSIZES",
            "GAPSIZES1", "GAPSIZES2" ) )
    
    ##-----------------------------------------------------------------------------------
    def GetLegend( self ):
        return UnalignedPair.GetLegend(self) + "\n" +\
               """# IDENTICAL:       number of identical positions
# TRANSITIONS:    number of transitions
# TRANSVERSIONS:  number of transversion
# NUMGAPS:        number of gaps
# LENGTH:         length of alignment
# ALIGNED:        number of aligned positions
# UNALIGNED:      number of unaligned positions
# ALIFROM1:       start of alignment in sequence1
# ALIEND1:        end of alignment in sequence1
# ALIGNED1:       alignment of sequence1
# ALIFROM2:       start of alignment in sequence2
# ALIEND2:        end of alignment in sequence2
# ALIGNED2:       alignment of sequence2
# BLOCKSIZES:     alignment, length of blocks
# GAPSIZES1:      length of gap sizes in sequence1
# GAPSIZES2:      length of gap sizes in sequence2"""

    ##-----------------------------------------------------------------------------------
    def GetMap( self ):
        """return map between the two segments."""
        if self.mAlignmentFrom1 and self.mAlignmentFrom2:
            map_a2b = alignlib.makeAlignmentVector()
            alignlib.AlignmentFormatEmissions( 
                self.mAlignmentFrom1, self.mAlignment1,
                self.mAlignmentFrom2, self.mAlignment2 ).copy( map_a2b )
            return map_a2b
        else:
            return None

    ##-----------------------------------------------------------------------------------
    def Align( self, method, anchor = 0, loglevel = 1 ):
        """align a pair of sequences.
        get rid of this and use a method class instead in the future
        """
        
        map_a2b = alignlib.makeAlignmentVector()
        s1 = "A" * anchor + self.mSequence1 + "A" * anchor
        s2 = "A" * anchor + self.mSequence2 + "A" * anchor    

        self.strand = "+"

        if method == "dialign":
            dialign = WrapperDialign.Dialign( self.mOptionsDialign )
            dialign.Align( s1, s2, map_a2b )
        elif method == "blastz":
            blastz = WrapperBlastZ.BlastZ( self.mOptionsBlastZ )
            blastz.Align( s1, s2, map_a2b )
            if blastz.isReverseComplement():
                self.strand = "-"
                self.mSequence2 = Genomics.complement( self.mSequence2 )

        elif method == "dialignlgs":
            dialignlgs = WrapperDialign.Dialign( self.mOptionsDialignLGS )
            dialignlgs.Align( s1, s2, map_a2b ) 
        elif method == "dba":
            dba = WrapperDBA.DBA()
            dba.Align( s1, s2, map_a2b )
        elif method == "clustal":
            raise NotImplementedError( "clustal wrapper needs to be updated")
            clustal = WrapperClustal.Clustal()
            clustal.Align( s1, s2, map_a2b )
        elif method == "nw":
            seq1 = alignlib.makeSequence( s1 )
            seq2 = alignlib.makeSequence( s2 )
            alignator = alignlib.makeAlignatorDPFull( alignlib.ALIGNMENT_GLOBAL,
                                                      gop=-12.0,
                                                      gep=-2.0 )
            alignator.align( map_a2b, seq1, seq2 )
        elif method == "sw":                        
            seq1 = alignlib.makeSequence( s1 )
            seq2 = alignlib.makeSequence( s2 )
            alignlib.performIterativeAlignment( map_a2b, seq1, seq2, alignator_sw, min_score_sw )
        else:
            ## use callback function
            method(s1, s2, map_a2b)

        if map_a2b.getLength() == 0:
            raise AlignmentError("empty alignment")

        if anchor:
            map_a2b.removeRowRegion( anchor + len(self.mSequence1) + 1, map_a2b.getRowTo() )
            map_a2b.removeRowRegion( 1, anchor)        
            map_a2b.removeColRegion( anchor + len(self.mSequence2) + 1, map_a2b.getColTo() )        
            map_a2b.removeColRegion( 1, anchor)
            map_a2b.moveAlignment( -anchor, -anchor )

        f = alignlib.AlignmentFormatExplicit( map_a2b, 
                                              alignlib.makeSequence( self.mSequence1),
                                              alignlib.makeSequence( self.mSequence2) )

        self.mMethod = method
        self.mAlignment = map_a2b
        self.mAlignedSequence1, self.mAlignedSequence2 = f.mRowAlignment, f.mColAlignment
        f = alignlib.AlignmentFormatEmissions( map_a2b )
        self.mAlignment1, self.mAlignment2 = f.mRowAlignment, f.mColAlignment
        self.mAlignmentFrom1 = map_a2b.getRowFrom()
        self.mAlignmentTo1 = map_a2b.getRowTo()        
        self.mAlignmentFrom2 = map_a2b.getColFrom()
        self.mAlignmentTo2 = map_a2b.getColTo()        
        self.mNumGaps, self.mLength = map_a2b.getNumGaps(), map_a2b.getLength()
        self.mAligned = self.mLength - self.mNumGaps

        self.SetPercentIdentity()
        self.SetBlockSizes()

    ##-----------------------------------------------------------------------------------
    def SetBlockSizes( self, gap_char = "-" ):
        """return the block-sizes in the alignment.
        """
        block_sizes = []
        left_gap_sizes = []
        right_gap_sizes = []    

        was_left_gap = self.mAlignedSequence1[0] == gap_char
        was_block = self.mAlignedSequence1[0] != gap_char and self.mAlignedSequence2[0] != gap_char
        size = 1

        for x in range(1,min(len(self.mAlignedSequence1), len(self.mAlignedSequence2))):
            is_left_gap  = self.mAlignedSequence1[x] == gap_char
            is_right_gap = self.mAlignedSequence2[x] == gap_char
            is_block = not is_left_gap and not is_right_gap
            if is_block and not was_block:
                if was_left_gap:
                    left_gap_sizes.append(size)
                    size = 0
                else:
                    right_gap_sizes.append(size)
                    size = 0
            elif not is_block and was_block:
                block_sizes.append(size)
                size = 0
            elif is_left_gap and not was_left_gap:
                right_gap_sizes.append(size)
                size = 0
            elif is_right_gap and was_left_gap:
                left_gap_sizes.append(size)
                size = 0
            elif is_left_gap and is_right_gap:
                raise "double gap"

            was_block = is_block
            was_left_gap = is_left_gap

            size += 1

        if was_block:
            block_sizes.append(size)
        else:
            if was_left_gap:
                left_gap_sizes.append(size)
            else:
                right_gap_sizes.append(size)

        self.mBlockSizes, self.mGapSizes1, self.mGapSizes2 = block_sizes, left_gap_sizes, right_gap_sizes

    ##-----------------------------------------------------------------------------------
    def SetPercentIdentity( self, gap_char = "-" ):
        """return number of idential and transitions/transversions substitutions
        in the alignment.
        """
        transitions   = ("AG", "GA", "CT", "TC")
        transversions = ("AT", "TA", "GT", "TG", "GC", "CG", "AC", "CA" )    

        nidentical = 0
        naligned = 0
        ndifferent = 0
        ntransitions = 0
        ntransversions = 0
        nunaligned = (self.mAlignmentFrom1 - self.mFrom1) +\
                     (self.mTo1 - self.mAlignmentTo1 ) +\
                     (self.mAlignmentFrom2 - self.mFrom2) +\
                     (self.mTo2 - self.mAlignmentTo2 )                      
        
        for x in range(min(len(self.mAlignedSequence1), len(self.mAlignedSequence2))):
            if self.mAlignedSequence1[x] != gap_char and \
               self.mAlignedSequence2[x] != gap_char:
                naligned += 1
                if self.mAlignedSequence1[x] == self.mAlignedSequence2[x]:
                    nidentical += 1
                else:
                    ndifferent += 1

                if (self.mAlignedSequence1[x] + self.mAlignedSequence2[x]) in transitions:
                    ntransitions += 1
                if (self.mAlignedSequence1[x] + self.mAlignedSequence2[x]) in transversions:
                    ntransversions += 1
            else:
                nunaligned += 1

        self.mIdentical, self.mTransitions, self.mTransversions, self.mAligned, self.mUnaligned = \
                         nidentical, ntransitions, ntransversions, naligned, nunaligned








################################################################################
#   MRC FGU Computational Genomics Group
#
#   $Id: fasta2counts.py 2781 2009-09-10 11:33:14Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Antonio changes up to 15/08/13:
- edits to documentation, modified header, added tags after Python, added under Usage, added under Example
- added code (commented)
 


fasta2counts.py - basic stats from collection of sequences
==========================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python, fasta, counts, contigs, residues, gaps

Purpose
-------

This script computes some basic counts from a fasta-formatted file. For each sequence in
the :term:`fasta` formatted and indexed input file it will output in a :term:`tsv` formatted table:

contig
   Sequence identifier
nresidues
   Number of residues
ngaps
   Number of gaps
nseqregions
   Number of ungapped regions
ngapregions
   Number of gapped regions
nA, nC, nG, nT, nN, nX, nO
   Number of A,C,T,G,N,X,O characters in sequence

Usage
-----

-g, --genome-file is a positional argument which requires an indexed fasta file created with index_fasta.py


Example::

#Download the example data here:

   wget ftp://ftp.ncbi.nih.gov/genomes/Bacteria/Pseudomonas_aeruginosa_PAO1_uid57945/NC_002516.fna

#Index your fasta file:

   python /ifs/devel/antoniob/cgat/scripts/index_fasta.py NC_002516_db NC_002516.fna 

#which outputs:

NC_002516_db.fasta
NC_002516_db.idx

#Run fasta2counts.py:

   python /ifs/devel/antoniob/cgat/scripts/fasta2counts.py --stdin=NC_002516.fna --log=fasta2counts_NC_002516.log --error=fasta2counts_NC_002516.error --stdout=fasta2counts_NC_002516.output --genome-file=NC_002516_db.fasta 

#Results should look like:

   cat fasta2counts_NC_002516.output 
contig	nresidues	ngaps	nseqregions	ngapregions	nA	nC	nG	nT	nN	nX	nO
gi|110645304|ref|NC_002516.2|	6264404	0	1	0	1056134	2102687	2066633	1038950	0	0	0

Type::

   python fasta2counts.py --help

for command line help.

'''

import os
import sys
import string
import re
import getopt
import tempfile
import time
import optparse
import math
import glob

import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta

def writeHeader( outfile ):
    outfile.write( "\t".join( ("contig",
                               "nresidues",
                               "ngaps",
                               "nseqregions",                                      
                               "ngapregions", 
                               "nA", "nC", "nG", "nT",
                               "nN", "nX", "nO" ) ) + "\n" )
    

    

if __name__ == "__main__":

#added ' usage = globals()["__doc__"]) '
    parser = E.OptionParser( version = "%prog version: $Id: fasta2counts.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"])

#added ' ..., requires indexed fasta file" ' 
    parser.add_option( "-g", "--genome-file", dest="genome_file", type="string",
                       help="filename with genome, requires indexed fasta file"  )

    parser.set_defaults(
        genome_file = None,
        gap_chars = 'xXnN',
        filename_total = None,
        )

    (options, args) = E.Start( parser )

    fasta = IndexedFasta.IndexedFasta( options.genome_file )
    contigs = fasta.getContigSizes( with_synonyms = False )

    totals = { 'A' : 0, 'C': 0, 'G' : 0, 'T' : 0, 'X' : 0, 'N' : 0 }
    nothers_total = 0
    nresidues_total = 0
    ngap_regions_total = 0
    nseq_regions_total = 0

    writeHeader( options.stdout )

    for contig in contigs.keys():
        
        subtotals = {}
        for x in totals.keys(): subtotals[x] = 0
        
        nothers = 0
        
        s = fasta.getSequence( contig, "+", 0, 0 )

        ngaps = 0
        
        was_gap = not (s[0] in options.gap_chars)
        
        ngap_regions = 0
        nseq_regions = 0

        x = 0
        xx = len(s)
        while x < xx:
            c = s[x].upper() 
            if c in subtotals:
                subtotals[c] += 1
            else:
                nothers += 1

            is_gap = c in options.gap_chars
            if is_gap:
                if not was_gap: ngap_regions += 1
            else:
                if was_gap: nseq_regions += 1
            was_gap = is_gap

            x += 1
            
        ngaps = subtotals['N'] + subtotals['X']
        
        options.stdout.write( "\t".join( map(str, (contig, len(s), ngaps,
                                                   nseq_regions,
                                                   ngap_regions,
                                                   subtotals['A'], subtotals['C'], subtotals['G'], subtotals['T'],
                                                   subtotals['N'], subtotals['X'], nothers ) ) ) + "\n" )
        

        for x in subtotals.keys(): totals[x] += subtotals[x]
        nothers_total += nothers
        nresidues_total += len(s)
        ngap_regions_total += ngap_regions
        nseq_regions_total += nseq_regions
        
    ngaps = totals['N'] + totals['X']

    if options.filename_total:
        outfile = open(options.filename_total, "w" )
        writeHeader( outfile )
        outfile.write( "\t".join( map(str, (len(contigs), nresidues_total, ngaps,
                                               nseq_regions_total,
                                               ngap_regions_total,
                                               totals['A'], totals['C'], totals['G'], totals['T'],
                                               totals['N'], totals['X'], nothers_total ) ) ) + "\n" )
        
        outfile.close()
        
    E.Stop()




################################################################################
#   Gene prediction pipeline 
#
#   $Id: WrapperBaseML.py 2765 2009-09-04 16:55:18Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
Wrapper for BaseML
==================

"""
import os, sys, string, re, tempfile, subprocess

import Genomics

class Error(Exception):
    """Base class for exceptions in this module."""
    pass

    def __str__(self):
        return str(self.message)

class ParsingError(Error):
    """Exception raised for errors while parsing

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class UsageError(Error):
    """Exception raised for errors while starting

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class BaseMLResult:
    def __init__(self):
        pass

class BaseML:
    
    """
    Parser and outputs configuration files like this::

                seqfile = brown.nuc
                treefile = brown.trees

                outfile = mlb       * main result file
                noisy = 9     * 0,1,2,3: how much rubbish on the screen
                verbose = 0   * 1: detailed output, 0: concise output
                runmode = 0   * 0: user tree;  1: semi-automatic;  2: automatic
                              * 3: StepwiseAddition; (4,5):PerturbationNNI -2: pairwise

                model = 4   * 0:JC69, 1:K80, 2:F81, 3:F84, 4:HKY85
                            * 5:T92, 6:TN93, 7:REV, 8:UNREST, 9:REVu; 10:UNRESTu

                Mgene = 0   * 0:rates, 1:separate; 2:diff pi, 3:diff kapa, 4:all diff

        *        ndata = 5
                clock = 0   * 0:no clock, 1:clock; 2:local clock; 3:CombinedAnalysis
            fix_kappa = 0   * 0: estimate kappa; 1: fix kappa at value below
                kappa = 5  * initial or fixed kappa

            fix_alpha = 0   * 0: estimate alpha; 1: fix alpha at value below
                alpha = 0.5   * initial or fixed alpha, 0:infinity (constant rate)
               Malpha = 0   * 1: different alpha's for genes, 0: one alpha
                ncatG = 5   * # of categories in the dG, AdG, or nparK models of rates
                nparK = 0   * rate-class models. 1:rK, 2:rK&fK, 3:rK&MK(1/K), 4:rK&MK

                nhomo = 0   * 0 & 1: homogeneous, 2: kappa for branches, 3: N1, 4: N2
                getSE = 0   * 0: don't want them, 1: want S.E.s of estimates
         RateAncestor = 1   * (0,1,2): rates (alpha>0) or ancestral states

           Small_Diff = 7e-6
            cleandata = 1  * remove sites with ambiguity data (1:yes, 0:no)?
        *        icode = 0  * (with RateAncestor=1. try "GC" in data,model=4,Mgene=4)
        *  fix_blength = -1  * 0: ignore, -1: random, 1: initial, 2: fixed
                method = 0  * 0: simultaneous; 1: one branch at a time
    """

    mOptions = {
        "noisy"         : 9,
        "verbose"       : 0,
        "runmode"       : 0, 
        "model"         : 4,     
        "Mgene"         : 0,
        "clock"         : 0,
        "fix_kappa"     : 0,
        "kappa"         : 5,
        "fix_alpha"     : 0, 
        "alpha"         : 0.5, 
        "Malpha"        : 0,
        "ncatG"         : 5,
        "nparK"         : 0,
        "nhomo"         : 0,
        "getSE"         : 0,
        "RateAncestor"  : 1,
        "Small_Diff"    : "7e-6",
        "cleandata"     : 1,
        "method"        : 0,
        }
    mExecutable = "baseml"

    def __init__( self ):

        self.mFilenameTree = None
        self.mFilenameSequences = None

    def GetOptions( self ):
        """return options in pretty format"""
        result = ["# Options for BaseML" ]
        for var in self.mOptions.keys():
            result.append("# %-40s: %s" % (var, self.mOptions[var]))
        return string.join(result, "\n")
        
    def SetOption( self, option, value ):
        self.mOptions[option] = value

    def WriteAlignment( self, alignment ):
        """write alignment in Phylip format."""

        outfile = open( self.mTempdir + "/" + self.mFilenameSequences, "w" )

        outfile.write( " %4i %i\n" % (len(alignment), len(alignment[0][1])) )
        for i, s in alignment:
            outfile.write("%s\n%s\n" % (i, s))

        outfile.close()

    def WriteTree( self, alignment ):
        """write tree to file."""

        outfile = open( self.mTempdir + "/" + self.mFilenameTree, "w" )

        if len(alignment) != 2:
            raise UsageError( "only 2 sequences right now!")

        outfile.write("(%s,%s);\n" % (alignment[0][0], alignment[1][0]))
        outfile.close()
        
    def WriteControlFile( self ):
        """write baseml.ctl"""
        
        outfile = open( self.mTempdir + "/baseml.ctl", "w" )

        if self.mFilenameSequences:
            outfile.write( "seqfile = %s\n" % (self.mFilenameSequences) )
        if self.mFilenameTree:
            outfile.write( "treefile = %s\n" % (self.mFilenameTree) )

        outfile.write( "outfile = %s\n" % (self.mFilenameOutput) )

        for o,a in self.mOptions.items():
            outfile.write( "%s = %s\n" % (o, str(a)))
            
        outfile.close()
    
    def Run( self, alignment, dump_result = 0 ):

        self.mTempdir = tempfile.mkdtemp()
        self.mFilenameSequences = "input"
        self.mFilenameTree = "tree"
        self.mFilenameOutput = "output"        
        self.mNumSequences = len(alignment)
        
        self.WriteAlignment( alignment )
        self.WriteTree( alignment )
        self.WriteControlFile()
        
        statement = "cd %s; %s;" % (self.mTempdir, self.mExecutable )

        p = subprocess.Popen( statement , 
                              shell=True, 
                              stdin=subprocess.PIPE, 
                              stdout=subprocess.PIPE, 
                              stderr=subprocess.PIPE, 
                              close_fds=True)

        (file_stdout, file_stdin, file_stderr) = (p.stdin, p.stdout, p.stderr)
        
        file_stdin.close()
        lines = file_stdout.readlines()
        lines_stderr = file_stderr.readlines()
        exit_code = file_stdout.close()
        file_stderr.close()
        file_stdout.close()

        lines = open( "%s/%s" % (self.mTempdir, self.mFilenameOutput), "r").readlines()

        if dump_result: print string.join(lines, "")
        os.system( "rm -rf %s" % self.mTempdir )
        
        return self.ParseResult( lines )

    def ParseResult( self, lines ):
        """parse BaseML output.
        """

        """
BASEML (in paml 3.14, January 2004)  input  HKY85 dGamma (ncatG=5)
ns = 2          ls = 60
# site patterns = 7
   15    7   18   16    1    2    1

sequence 1                           AGCTGCT
sequence 2                           ....ATC




Frequencies..
                                    T      C      A      G
sequence 1                     0.2833 0.3333 0.2500 0.1333
sequence 2                     0.3000 0.3167 0.2667 0.1167

Average                        0.2917 0.3250 0.2583 0.1250

# constant sites:     56 (93.33%)
ln Lmax (unconstrained) = -93.644144

Distances:HKY85 (kappa)  (alpha set at 0.50)
This matrix is not used in later m.l. analysis.

sequence 1
sequence 2         0.0823(999.0000)

====================================================================
BASEML (in paml 3.14, January 2004)  input  JC69 dGamma (ncatG=5)
ns = 2          ls = 60
# site patterns = 7
   15    7   18   16    1    2    1

sequence 1                           AGCTGCT
sequence 2                           ....ATC




Frequencies..
                                    T      C      A      G
sequence 1                     0.2833 0.3333 0.2500 0.1333
sequence 2                     0.3000 0.3167 0.2667 0.1167

Average                        0.2917 0.3250 0.2583 0.1250

# constant sites:     56 (93.33%)
ln Lmax (unconstrained) = -93.644144

Distances: JC69 (alpha set at 0.50)
This matrix is not used in later m.l. analysis.

sequence 1
sequence 2         0.0767
BASEML (in paml 3.14b, May 2005)  input  HKY85 dGamma (ncatG=5)
ns = 2          ls = 314
# site patterns = 16
   77   19   61   66    4    3   20   25    7    6    6    8    5    3    2
   2
seq1                                 AGGTTAACTG CCATGC
seq2                                 .A..GTG.AC ATCCTG
Frequencies..
T      C      A      G
seq1                           0.2548 0.1306 0.3344 0.2803
seq2                           0.2516 0.1242 0.3471 0.2771
Homogeneity statistic: X2 = 0.00044 G = 0.00044
Average                        0.2532 0.1274 0.3408 0.2787
# constant sites:    229 (72.93%)
ln Lmax (unconstrained) = -672.510333
Distances:HKY85 (kappa)  (alpha set at 0.50)
This matrix is not used in later m.l. analysis.
seq1
seq2               0.5990( 4.7311)
TREE #  1:  (1, 2);  MP score: 85.00
lnL(ntime:  2  np:  4):   -679.913467   +0.000000
  3..1     3..2
    0.18769  0.16256  3.25628 999.00000
tree length =   0.35024
(seq1, seq2);
(seq1: 0.187685, seq2: 0.162558);
Detailed output identifying parameters
Parameters (kappa) in the rate matrix (HKY85) (Yang 1994 J Mol Evol 39:105-111):
  3.25628
alpha (gamma, K=5) = 999.00000
r:   0.95611  0.98294  0.99967  1.01660  1.04468
f:   0.20000  0.20000  0.20000  0.20000  0.20000

################
Explanation
TREE #  1:  (1, 2);  MP score: 85.00
lnL(ntime:  2  np:  4):   -679.913467   +0.000000
    3..1     3..2   # branches
    0.18769  0.16256  3.25628 999.00000 # branch1 branch2 kappa alpha
tree length =   0.35024                 # distance
(seq1, seq2);
(seq1: 0.187685, seq2: 0.162558);       # tree format
Detailed output identifying parameters  # 
Parameters (kappa) in the rate matrix (HKY85) (Yang 1994 J Mol Evol 39:105-111):
  3.25628
alpha (gamma, K=5) = 999.00000         # gamma distribution, here: five bins, uniform
r:   0.95611  0.98294  0.99967  1.01660  1.04468
f:   0.20000  0.20000  0.20000  0.20000  0.20000

If alpha is fixed, there are only three values for alpha.

"""
        result = BaseMLResult()        
        # chop
        lines = map(lambda x: x[:-1], lines )
        lines = filter( lambda x: x != "", lines)

        result.mVersion, result.mModel = re.search("BASEML \(in (.+)\)\s+input\s+(\S.+)", lines[0]).groups()

        try:
            if result.mVersion == "paml 3.14, January 2004":
                result.mSitePatterns = map(int, re.split("\s+", lines[0].strip()))
                while not re.match("^Frequencies", lines[0] ): del lines[0]
                result.mFrequencies = {}
                result.mMatrix = {}
                for x in range(self.mNumSequences):
                    id = lines[0][:30].strip()
                    result.mMatrix[id] = {}
                    result.mFrequencies[id] = map(float, re.split( "\s+", lines[0][30:].strip()))
                    del lines[0]
                result.mFrequencies['average'] = map(float, re.split( "\s+", lines[0][30:].strip()))
                result.mConstantSites, result.mConstantSitesPercent = \
                                       re.search( "# constant sites:\s+(\S+)\s\((.+)%\)", lines[0] ).groups()
                result.mLmax  = re.search( "ln Lmax \(unconstrained\) = (\S+)", lines[0] ).groups()
                previous_ids = []
                for x in range(self.mNumSequences):
                    id = lines[0][:17].strip()
                    rest = lines[0][17:].strip()
                    if rest:
                        values = re.findall("[\d.+-]+", rest)                    
                        if self.mOptions['model'] == 4:
                            values = map(float, values[::2])
                        else:
                            values = map(float, values )
                        for i in range(len(previous_ids)):
                            result.mMatrix[previous_ids[i]][id] = values[i]
                            result.mMatrix[id][previous_ids[i]] = values[i]                
                    previous_ids.append(id)
            elif result.mVersion == "paml 3.14b, May 2005":
                # print string.join(lines,"\n")
                while not re.match("^# site patterns", lines[0] ): del lines[0]
                del lines[0]
                result.mSitePatterns = map(int, re.split("\s+", lines[0].strip()))

                result.mFrequencies = {}
                result.mMatrix = {}
                while not re.match("^Frequencies", lines[0] ): del lines[0]
                del lines[:2]
                for x in range(self.mNumSequences):
                    id = lines[0][:30].strip()
                    result.mMatrix[id] = {}
                    result.mFrequencies[id] = map(float, re.split( "\s+", lines[0][30:].strip()))
                    del lines[0]
                result.mX2, result.mG = map(float, re.search( "Homogeneity statistic: X2 = (\S+) G = (\S+)", lines[0] ).groups())
                del lines[0]
                result.mFrequencies['average'] = map(float, re.split( "\s+", lines[0][30:].strip()))
                del lines[0]
                result.mConstantSites, result.mConstantSitesPercent = \
                                       map(float, re.search( "# constant sites:\s+(\S+)\s\((.+)%\)", lines[0] ).groups())

                while lines and not re.match("^TREE", lines[0] ):
                    del lines[0]

                result.mBranch1, result.mBranch2, result.mKappa, result.mAlpha = 0.0, 0.0, 0.0, 0.0                                                        
                ## sometimes, nothing after tree follows,
                ## for example, if there are no variable sites, no estimation possible
                ## take only branch lengths from the next line, because alpha and kappa might or might
                ## not be present
                del lines[:3]
                if lines:
                    data = map(float, re.split("\s+", lines[0].strip()))

                    result.mBranch1, result.mBranch2 = data[:2]

                    if self.mOptions["fix_alpha"] == 0:
                        self.mKappa, self.mAlpha = data[2:4]
                    else:
                        self.mKappa = data[2]
                    
                result.mDistance = result.mBranch1 + result.mBranch2
                keys = result.mMatrix.keys()
                result.mMatrix[keys[0]][keys[1]] = result.mDistance
                result.mMatrix[keys[1]][keys[0]] = result.mDistance            
            else:
                raise ParsingError( "unknown BaseML version" )

        except Exception, inst:
            raise ParsingError( str(inst) + ":" + lines[0] )
            
        return result

if __name__ == "__main__":
    
    baseml = BaseML()

    alignment = ( ("sequence 1", "AAGCTTCACCGGCGCAGTCATTCTCATAATCGCCCACGGACTTACATCCTCATTACTATT"),
                  ("sequence 2", "AAGCTTCACCGGCGCAATTATCCTCATAATCGCCCACGGACTTACATCCTCATTATTATT") )
                  
    baseml.Run( alignment )
    
        



################################################################################
#   Gene prediction pipeline 
#
#   $Id: Cluster.py 2784 2009-09-10 11:41:14Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
Cluster.py - module for running a job in parallel on the cluster
================================================================
"""

import sys, os, subprocess

class Error(Exception):
    """Base class for exceptions in this module."""
    pass

    def __str__(self):
        return str(self.message)

class ClusterError(Error):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class Cluster:

    def __init__(self, options):
        
        try:
            self.mPriority = options.cluster_priority
        except AttributeError:
            self.mPriority = None

        try:
            self.mQueue = options.cluster_queue
        except AttributeError:
            self.mCluster = None

        try:
            self.mNumJobs = options.cluster_num_jobs
        except AttributeError:
            self.mNumJobs = None
            
    def addJob( self, statement ):
        
        self.mJobs.append( statement )

    def runJobs( self ):
        """run jobs.

        This uses forking which is not without problems:

        * too many open files.

        Thus, running in chunks.
        """
        
        self.mResults = []

        if self.mNumJobs:
            chunks = range( 0, len(self.mJobs), self.mNumJobs)
            chunks.append( len(self.mJobs) )
        else:
            chunks = [0, len(self.mJobs) ]

        for x in range(len(chunks)-1):
            
            pids = []
            
            for job in self.mJobs[chunks[x]:chunks[x+1]]:

                pid = os.fork()

                if not pid:
                    os.nice( 20 )
                    ## TODO: deal with return values
                    out, err = self.__runJob( job )
                    self.mResults.append( (out, err) )
                    break
                else:
                    pids.append(pid)
                
            for pid in pids:
                os.waitpid(pid,0)

        return self.mResults

    def runJob( self, job ):
        """submit job to queue.

        This subroutine should only exit if the job completed.
        """
        s = subprocess.Popen( self.getStatement(job),
                              shell = True,
                              stdin = subprocess.PIPE,
                              stdout = subprocess.PIPE,
                              stderr = subprocess.PIPE,
                              cwd = os.curdir,
                              close_fds = True)                              

        (out, err) = s.communicate()

        if s.returncode != 0:
            raise ClusterError ("Error while running job %s: %s\n%s\n" % (job, out, err))

        return out, err

    def getStatement( self, job):
        return job

    def setPriority( self, priority ):
        self.mPriority = priority

    def setQueue( self, queue):
        self.mQueue = queue
    
class ClusterSGE(Cluster):

    def __init__(self, *args):
        
        Cluster.__init__(self, *args)

        if not self.mQueue:
            self.mQueue = "medium_jobs.q"
        if not self.mPriority:
            self.mPriority = -10

        self.mExportedVariables = ( "PYTHONPATH", "LD_LIBRARY_PATH", "PATH", "BLASTMAT", "DIALIGN2_DIR", "LANG", "WISECONFIGDIR" )

        self.mVariables = []
        for x in self.mExportedVariables:
            try:
                self.mVariables.append( ("-v %s='%s'" % (x, os.environ[x]) ) )
            except KeyError:
                pass
        self.mVariables = " ".join( self.mVariables )

        self.mOptions ="-l arch=lx24-x86"
        self.mName = os.path.basename(sys.argv[0])
        if not self.mName: self.mName = "generic"

    def getStatement(self, job):

        return "qrsh -now n -cwd -q %s -p %i -N %s %s %s %s " % ( self.mQueue,
                                                                  self.mPriority,
                                                                  self.mName,
                                                                  self.mVariables,
                                                                  self.mOptions,
                                                                  job )
    
        



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
rename_links.py - rename all links in a directory
=================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script reads links in a directory that match a pattern
and changes the destination into a new pattern keeping 
name of the link unchanged.

Usage
-----

Example::

   python rename_links.py --help

Type::

   python rename_links.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import fnmatch

import CGAT.Experiment as E

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: rename_links.py 2782 2009-09-10 11:40:29Z andreas $",
                                    usage = globals()["__doc__"])

    parser.add_option( "-r", "--recursive", dest="recursive", action="store_true",
                       help="proceed recursively through directories." )

    parser.add_option( "-d", "--dry-run", dest="dry_run", action="store_true",
                       help="dry run, do not change anything." )

    parser.add_option( "-o", "--old-pattern", dest="old_pattern", type="string",
                       help="pattern to be replaced.")

    parser.add_option( "-n", "--new-pattern", dest="new_pattern", type="string",
                       help="pattern to be inserted.")

    parser.set_defaults(
        recursive = False,
        dry_run = False,
        old_pattern = None,
        new_pattern = None,
        )

    (options, args) = E.Start( parser )

    if options.old_pattern == None or options.new_pattern == None:
        raise "please specify both an old and a new pattern."

    rx_old = re.compile(options.old_pattern)

    def changeLink( fn, options ):
        old_target = os.path.realpath( fn )
        if not re.search( options.old_pattern, old_target): return 0
        new_target = re.sub( options.old_pattern, options.new_pattern, old_target )

        cmds = []
        cmds.append('rm -f %s' % (fn))
        
        cmds.append('ln -fs %s %s' % (new_target, fn))

        if options.loglevel >= 1 or options.dry_run:
            for cmd in cmds:
                options.stdlog.write("%s\n" % cmd )
        
        if not options.dry_run:
            for cmd in cmds:
                os.system( cmd )

        return 1

    ninput, nvisited, nchanged = 0, 0, 0
    for directory in args:
        ninput += 1
        E.info( "processing directory %s" % directory )
            
        for root, directories, files in os.walk(directory):
            for f in files + directories:
                fn = os.path.join( root, f )
                if not os.path.islink(fn): continue
                nchanged += changeLink( os.path.abspath(fn), options )
                nvisited += 1

    if options.stdlog >= 1:
        options.stdlog.write("# ndirs=%i, nvisited=%i, nchanged=%i\n" % (ninput, nvisited, nchanged))
    
    E.Stop()






################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: gtf2alleles.py 2886 2010-04-07 08:47:46Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
gtf2alleles.py - predict effects of variants on gene set
========================================================

:Author: Andreas Heger
:Release: $Id: gtf2alleles.py 2886 2010-04-07 08:47:46Z andreas $
:Date: |today|
:Tags: Python

Purpose
-------

build from a gene set in gtf format an allelic gene set in which
each transcript is split into separate alleles. 

Alleles are built from variants stored in an sqlite database (see the
options ``--database`` and ``--tablename``).

Alleles are built from transcripts in a gene set (*--filename-exons*).

Caveats
-------

* no phasing - heterozygous variants are assigned to the first or second
  allele simply based on the input code.

Usage
-----

Type::

   python <script_name>.py --help

for command line help.

Code
----

""" 

import os
import sys
import re
import optparse
import collections
import sqlite3

import numpy
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Genomics as Genomics
import CGAT.GTF as GTF
import CGAT.Blat as Blat
import CGAT.Variants as Variants
import alignlib
import pysam

Allele = collections.namedtuple( 'Allele',
                                 '''cds, 
                                    peptide, 
                                    nexons,
                                    cds_starts,
                                    exon_starts,
                                    frames,
                                    is_nmd_knockout,
                                    is_splice_truncated,
                                    is_stop_truncated,
                                    nframeshifts,
                                    ncorrected_frameshifts,
                                    nuncorrected_frameshits,
                                    peptide_first_stop,
                                    peptide_len,
                                    cds_first_stop,
                                    cds_len,
                                    reference_first_stop_start,
                                    reference_first_stop_end,
                                    cds_original_len,
                                    nsplice_noncanonical,
                                 ''')


class VariantGetter(object):
    '''base class for objects returning variants.'''
    pass

class VariantGetterSqlite(VariantGetter):
    '''retrieve variants from an sqlite table in pileup format.'''

    def __init__(self, dbname, tablename ):
        self.dbname = dbname
        self.tablename = tablename
        
        self.dbhandle = sqlite3.connect( dbname )

        self.statement = '''SELECT 
                   pos, reference, genotype 
                   FROM %(tablename)s
                   WHERE contig = '%(contig)s' AND 
                   pos BETWEEN %(start)s and %(end)s
                '''
        
    def __call__( self, contig, start, end ):

        cc = self.dbhandle.cursor()
        tablename = self.tablename
        cc.execute( self.statement % locals() )
        variants = map( Variants.Variant._make, cc.fetchall())
        cc.close()
        return variants

class VariantGetterPileup(VariantGetter):
    '''retrieve variants from file in pileup format.'''

    def __init__(self, filename ):
        self.tabix = pysam.Tabixfile( filename )
        self.contigs = set(self.tabix.contigs)

    def __call__(self, contig, start, end ):

        variants = []
        if contig not in self.contigs: return []

        for line in self.tabix.fetch( contig, start, end ):
            data = line[:-1].split()
            contig, pos, reference, genotype = data[:4]
            # fix 1-ness
            pos = int(pos) - 1
            variants.append( Variants.Variant._make( (pos, reference, genotype) ) )
        return variants
    
class VariantGetterVCF( VariantGetter ):
    '''retrieve variants from tabix indexed vcf file.'''
    
    def __init__(self, filename, sample ):
        self.sample = sample
        self.vcf = pysam.VCF()
        self.vcf.connect( filename )

        if sample not in self.vcf.getsamples():
            raise KeyError( "sample %s not vcf file" % sample )
        
    def __call__(self, contig, start, end ):
        
        variants = []
        s = self.sample

        try:
            iter = self.vcf.fetch( contig, start, end )
        except ValueError:
            # contigs not in variants, ignore
            return variants

        for row in iter:
            result = pysam.Pileup.vcf2pileup( row, s )
            if not result: continue
            variants.append( Variants.Variant._make( (result.pos, result.reference_base, result.genotype) ))

        return variants

def collectExonIntronSequences( transcripts, fasta ):
    '''collect all the wild type sequences for exons and introns
    
    exons and introns are indexed by their respective positions.

    The function changes coordinates in ``gtfs`` to reverse coordinates.
    '''

    contig = transcripts[0][0].contig 
    strand = transcripts[0][0].strand
    lcontig = fasta.getLength( contig )

    all_exons, all_introns = {}, {}
    for exons in transcripts:
        for exon in exons:
            exon.invert( lcontig )
            start, end = exon.start, exon.end
            key = start, end
            if key not in all_exons:
                all_exons[key] = fasta.getSequence( contig, strand, start, end ).lower()
                    
        intron_intervals = GTF.toIntronIntervals(exons)
        for start, end in intron_intervals:
            key = start, end
            if key not in all_introns:
                all_introns[key] = fasta.getSequence( contig, strand, start, end ).lower()
    return all_exons, all_introns


def buildCompactVariantSequences( variants, sequences ):
    '''build variant sequences by inserting ``variants`` into ``sequences``.

    The original frame of the sequence is maintained by
    converting the input sequence to a list. Each entry
    in the list corresponds to a position in a wild type.

    The wild type (WT) sequence is lower case

    SNP:                    variant (ambiguity codes for variants)
    homozygous insertion:   upper-case bases after lower-case (WT) base
    heterozygous insertion: lower-case bases after lower-case (WT) base
    homozygous deletion:    empty fields
    heterozygous deletion:  "-" after lower-case (WT) base

    returns a dictionary of lists.
    '''

    result = {}
    for key, sequence in sequences.iteritems():
        variant_seq = list( sequence.lower() )
        start, end = key
        
        # get all variants that overlap with sequences
        for var_start, var_end, values in variants.find( start, end ):
            reference, action, has_wildtype, variantseqs = values

            is_homozygous = len(variantseqs) == 1 and not has_wildtype
            
            rel_start, rel_end = var_start - start, var_end - start
            startoffset = max(0, start-var_start)
            endoffset = max(0,var_end-end)
            
            if action == "=":
                assert rel_start >= 0
                assert sequence[rel_start].upper() == reference, \
                    'reference base mismatch: expected %s, got %s at %i-%i' % \
                    (sequence[rel_start].upper(), reference, 
                     var_start, var_end)

                if is_homozygous:
                    variant_seq[rel_start] = variantseqs[0]
                else:
                    variant_seq[rel_start] = Genomics.resolveReverseAmbiguousNA( "".join(variantseqs) )
                    
            elif action == "-":

                xstart, xend = max(0,rel_start), min(len(sequence),rel_end)

                for variant in variantseqs:
                    # truncated for variants of unequal lengths (-AA/-AAA)
                    refseq = sequence[xstart:xend].upper()[:len(variant)]

                    assert refseq == variant[startoffset:len(variant) - endoffset], \
                    'reference base mismatch at deletion: expected %s %s %s, got %s[%i:%i] at %i-%i (%i-%i), action=%s' % \
                    (sequence[xstart-10:xstart],
                     refseq, 
                     sequence[xend:xend+10],
                     variant, startoffset, len(variant)-endoffset,
                     var_start, var_end, start, end, 
                     action)

                    l = len(variant) - startoffset - endoffset

                    if is_homozygous:
                        variant_seq[xstart:xend] = [""] * l
                    else:
                        for x in range(xstart,xend):
                            if variant_seq[x].endswith("-"):
                                assert not has_wildtype
                                variant_seq[x] = ""
                            else:
                                variant_seq[x] += "-"

            elif action == "+":

                if is_homozygous:
                    variant_seq[rel_start] += variantseqs[0].upper()
                else:
                    if has_wildtype:
                        variant_seq[rel_start] += variantseqs[0].upper()
                    else:
                        # merge indels like +AAA/+AA
                        a,b = variantseqs
                        if a.startswith(b):
                            variant_seq[rel_start] += b.upper() + a[len(b):].lower()
                        elif b.startswith(a):
                            variant_seq[rel_start] += a.upper() + b[len(a):].lower()
                        else:
                            raise ValueError("don't know how to encode variant: %s" % variantseqs)

        result[(start,end)] = variant_seq
    return result

def buildVariantSequences( indexed_variants, sequences ):
    '''build variant sequences by inserting ``variants`` into ``sequences``.

    For each sequence, two alleles are returned. Both alleles are initialized
    as wildtype sequences. In the absence of any phasing information, variants 
    are preferably added to the second allele, such that the wild-type status
    of the first allele is preserved as much as possible

    returns a dictionary of lists.
    '''

    result = {}
    for key, sequence in sequences.iteritems():

        feature_start, feature_end = key

        variants = [ (x,y,) + z for (x,y,z) in indexed_variants.find( feature_start, feature_end )]
        allele1, allele2 = Variants.buildAlleles( sequence, 
                                                  variants, 
                                                  reference_start = feature_start )

        result[(feature_start,feature_end)] = (allele1,allele2)

    return result

def buildCDSSequence( transcript, exons ):
    '''build transcript sequences from exons.'''

    cds = []
    for exon in transcript:
        cds.append( exons[exon.start,exon.end] )
    return "".join(cds)

def buildAlleles( transcript, 
                  variant_exons, 
                  variant_introns,
                  reference_exons,
                  reference_introns,
                  offsets,
                  is_seleno = False,
                  frameshiftsize = 5,
                  reference_coordinates = False):
    '''reconstitute variant transcript sequences from
    variant exons and introns. This method returns alleles
    as they are estimated to look like taking into account
    splice effects and nonsense mediated decay.

    Transcripts are build in the following way:
    
    1. The exonic sequences are examinated for stop codons.

    If they contain a stop-codon in the last exon, the sequence 
    will be truncated at the stop. If a stop-codon exists
    in exons before the last codon, the allele is thought to
    have been knocked out due to NMD and it is set to 0.

    2. If a splice-site is disrupted, the transcript is 
    truncated (it is not extended to the first stop, as
    splicing might be recovered).

    If ``variant_exons`` and ``variant_introns`` are set to
    None, the two wildtype alleles will be returned.

    Scenarios that this method ignores:
       1. insertion after 3' splice site are counted as belonging
       to the intron and not the exon, though they could be either.
       2. insertion before the 5' splice are counted as belonging
       to the exon and not the intron.

    If ``reference_coordinates`` is set to true, exon coordinates are
    taken from the reference. This ensures, that the base that is
    derived from the same reference sequence gets the same coordinate
    in the two alleles. The alternative is allelic coordinates, that
    take into account intron size changes. For example, the reference
    has two exons, (100,200) and (300,400). Due to variants, the
    first allele has now an intron size of 110. Thus, in allelic coordinates,
    the coordinates are ((100,200), (310,410)) and ((100,200),(300,400)),
    while in reference coordinates, both alleles whould have the same
    coordinates. Note that due to insertions/deletion, the coordinates
    might change within an exon, too.

    returns two alleles
    '''

    result = []

    def _buildAllele( allele_id, 
                      transcript, exons, 
                      introns, offsets, 
                      virtual_coordinates = False,
                      reference_exons = None ):

        def _getOffset( pos, offsets):
            x = 0
            while x < len(offsets) and offsets[x][0] <= pos: x += 1
            x -= 1
            if x >= 0: return offsets[x][1]
            else: return 0

        def _sumIndels( ss ):
            '''sum indels within ss'''
            c = 0
            for s in ss:
                c += len(s) - 1 
            return c

        def _getEndOffsets( ss ):
            '''get the offset at exons due to deletions at
            start/end of exon.'''
            l = len(ss)
            x = 0
            while x < l and ss[x] == "": x += 1
            start_offset = x

            x = l-1
            while x >= 0 and ss[x] == "": x -= 1
            if x >= 0: 
                return start_offset, (l-1) - x
            else:
                return start_offset, 0

        def _addCds2Reference( map_cds2reference,
                               cds_start,
                               cds_seq, 
                               reference_start ):
            '''add cds to reference'''
            c, r = cds_start, reference_start
            for x in cds_seq:
                l = len(x)
                if l == 0: 
                    r += 1
                else:
                    map_cds2reference.addPair( c, r )
                    c += l
                    r += 1
        # counts
        is_splice_truncated = False
        is_nmd_knockout = False
        is_stop_truncated = False
        nuncorrected_frameshifts = 0
        ncorrected_frameshifts = 0
        nframeshifts = 0
        nsplice_noncanonical = 0
        reference_first_stop_start = -1
        reference_first_stop_end = -1

        # map between the new cds sequence and the reference
        # sequence
        map_cds2reference = alignlib.makeAlignmentBlocks()

        ###################################################
        # process first exon
        exon = transcript[0]
        transcript_id = exon.transcript_id

        # collect offset for exon.start
        genome_start = exon.start
        genome_start += _getOffset( genome_start, offsets )
        lcds, cds = 0, []
        cds_starts = [0]

        # still need to deal with deletions of first base:
        exon_starts = [genome_start]
        exon_key = (exon.start, exon.end)
        exon_sequence = exons[exon_key] 
        exon_seq = "".join(exon_sequence) 
        
        cds.append(exon_seq)
        _addCds2Reference( map_cds2reference, 
                           lcds,
                           exon_sequence,
                           exon.start )
        lcds = len(exon_seq)

        if len(exon_seq) != exon.end - exon.start:
            nframeshifts += 1

        # add first exon to genome position
        genome_pos = genome_start + len(exon_seq)
        last_end = exon.end 

        # correct for deletions at start/end of exon
        start_offset, end_offset = _getEndOffsets( exon_sequence )

        # length of original transcript
        loriginal = sum( [x.end - x.start for x in transcript] )

        if E.global_options.loglevel >= 8:
            print "%i: exon_indels (%i-%i):" % (allele_id, exon.start, exon.end)
            for x,c in enumerate(exons[exon_key]):
                if len(c) != 1: print x + exon.start, ":%s:" % c 
            print
            print exons[exon_key]
            print "genome_pos=", genome_pos, \
                ",exon=%i-%i" % (genome_pos, genome_pos + len(exon_seq)), \
                ", len(exon_seq)=", len(exon_seq), \
                ", len(exon)=", exon.end - exon.start, \
                ", offsets=%i,%i," % (start_offset, end_offset), \
                ", offset at start=",_getOffset( exon.start, offsets), \
                ", offset at end=",_getOffset( exon.end, offsets)
            

        for exon in transcript[1:]:

            last_exon_sequence = exon_sequence
            last_start_offset, last_end_offset = start_offset, end_offset

            # get the next intron/exon parameters
            exon_key = (exon.start, exon.end)
            exon_sequence = exons[exon_key]
            start_offset, end_offset = _getEndOffsets( exon_sequence )
            intron_key = (last_end, exon.start)
            
            if last_end == exon.start:
                # catch empty introns
                intron_sequence = []
                intron_key = None
            else:
                intron_sequence = introns[intron_key] 

            intron_seq = "".join( intron_sequence )

            ###################################################
            ###################################################
            ###################################################
            # add preceding intron
            new_exon = True

            if len(intron_seq) > frameshiftsize:
                
                intron_name, intron_seq5, intron_seq3 = Genomics.GetIntronType( intron_seq )
                if intron_name == "unknown":
                    if intron_seq[:2].islower() and intron_seq[-2:].islower():
                        E.debug( "%s: transcript has unknown splice signal - kept because not a variant: %s: %s:%s" % \
                                     (transcript_id, intron_name, intron_seq5, intron_seq3 ) )
                        nsplice_noncanonical += 1
                    else:
                        is_splice_truncated = True
                        E.debug( "%s: transcript has splice truncated allele: %s: %s:%s" % \
                                     (transcript_id, intron_name, intron_seq5, intron_seq3 ) )
                        break
                # start a new exon
                cds_starts.append( lcds )
                
            else:
                # treat as frameshifting intron
                #
                # frame-shifting introns are checked if they are
                # fixed by indels either in the intron itself or
                # the terminal exon sequence. To this end, the effective
                # size of the intron is computed:
                # effective size of intron = 
                # indels at terminal x bases at previous exon
                # + size of intron
                # + indels at terminal x bases at next exon 
                effective_intron_size = len(intron_seq)
                previous_indels = _sumIndels( last_exon_sequence[max(0,-frameshiftsize):] )
                next_indels = _sumIndels( exon_sequence[:frameshiftsize] )
                effective_intron_size += previous_indels + next_indels

                if previous_indels + next_indels == 0 and len(intron_seq) % 3 == 0:
                    has_stop = "X" in Genomics.translate(intron_seq.upper(), 
                                                         is_seleno=is_seleno )
                else:
                    has_stop = False

                if effective_intron_size % 3 == 0 and not has_stop:
                    E.debug( "%s: fixed frame-shifting intron %i-%i of size %i (size:%i, indels:%i,%i)" % \
                                 (transcript_id, last_end, exon.start,
                                  effective_intron_size, 
                                  len(intron_seq),
                                  previous_indels, next_indels,))
                                  
                    # add to previous exon
                    cds.append( intron_seq )
                    lcds += len(intron_seq)
                    ncorrected_frameshifts += 1
                    new_exon = False
                else:
                    E.debug( "%s: could not fix frame-shifting intron %i-%i of size %i (size:%i, indels:%i,%i, has_stop=%i)" % \
                                 (transcript_id, last_end, exon.start,
                                  effective_intron_size, 
                                  len(intron_seq),
                                  previous_indels, next_indels,
                                  has_stop))

                    nuncorrected_frameshifts += 1
                    # start a new exon
                    cds_starts.append( lcds )

            if E.global_options.loglevel >= 8:
                print "%i: intron_indels (%i-%i):" % (allele_id, last_end, exon.start)
                if intron_key:
                    for x,c in enumerate(introns[intron_key]):
                        if len(c) != 1: print x + last_end, ":%s:" % c 
                    print
                    print introns[intron_key]
                    print "genome_pos=", genome_pos, \
                        ",intron=%i-%i" % (genome_pos, genome_pos + len(intron_seq)), \
                        ", len(intron_seq)=", len(intron_seq), \
                        ", len(intron)=", exon.start-last_end, \
                        ", offset at start=",_getOffset( last_end, offsets), \
                        ", offset at end=",_getOffset( exon.start, offsets)
                else:
                    print "empty intron"

            genome_pos += len(intron_seq)
            
            # assertion - check if genomic coordinate of intron is consistent with offset
            test_offset =_getOffset( exon.start, offsets )
            is_offset = genome_pos - exon.start
            assert is_offset == test_offset, "intron offset difference: %i != %i" % (is_offset, test_offset)

            ###################################################
            ###################################################
            ###################################################
            # add the exon
            exon_seq = "".join(exon_sequence)
            cds.append( exon_seq )

            if len(exon_seq) != exon.end - exon.start:
                nframeshifts += 1

            if new_exon:
                if reference_coordinates:
                    exon_starts.append( exon.start + start_offset )
                else:
                    exon_starts.append( genome_pos )

            _addCds2Reference( map_cds2reference, 
                               lcds, 
                               exon_sequence, 
                               exon.start )

            lcds += len(exon_seq)
            last_end = exon.end

            if E.global_options.loglevel >= 8:
                print "%i: exon_indels (%i-%i):" % (allele_id, exon.start, exon.end)
                for x,c in enumerate(exons[exon_key]):
                    if len(c) != 1: print x + exon.start, ":%s:" % c 
                print
                print exons[exon_key]
                print "genome_pos=", genome_pos, \
                    ",exon=%i-%i" % (genome_pos, genome_pos + len(exon_seq)), \
                    ", len(exon_seq)=", len(exon_seq), \
                    ", len(exon)=", exon.end - exon.start, \
                    ", offsets=%i,%i," % (start_offset, end_offset), \
                    ", offset at start=",_getOffset( exon.start, offsets), \
                    ", offset at end=",_getOffset( exon.end, offsets)

            genome_pos += len(exon_seq)

            test_offset =_getOffset( exon.end, offsets )
            is_offset = genome_pos - exon.end
            assert is_offset == test_offset, "exon offset difference: %i != %i" % (is_offset, test_offset)

        cds = "".join( cds )
        assert lcds == len(cds)

        # fix incomplete codons at the end of the sequence
        if lcds % 3 != 0:
            offset = lcds % 3
            cds=cds[:-offset]
        
        # add frame correction for transcripts that do not start at frame=0
        start_frame = (3 - (int(transcript[0].frame) % 3)) % 3

        # n are ignored (? in sequence to deal with genes like Muc2)
        peptide = Genomics.translate( "n" * start_frame + cds, 
                                      is_seleno = is_seleno, 
                                      prefer_lowercase = False,
                                      ignore_n = True )

        # find the first stop codon
        if start_frame != 0:
            # ignore first, potentially incomplete base
            pep_first_stop = peptide.upper().find( "X", 1 )
        else:
            pep_first_stop = peptide.upper().find( "X" )

        E.debug("%s: translated peptide = %s, first stop at %i" % (transcript_id, peptide, pep_first_stop))

        peptide = peptide.replace( "?", "x" )

        if E.global_options.loglevel >= 8:
            E.debug( "peptide=%s" % peptide )
            E.debug( "cds=%s" % cds )

        E.debug( "%s: start_frame=%i, first stop at %i/%i" % (transcript_id, 
                                                              start_frame,
                                                              pep_first_stop, 
                                                              len(peptide)))


        lpeptide, lcds = len(peptide), len(cds)

        # check for non-sense mediated decay
        if pep_first_stop != -1:
            cds_first_stop = pep_first_stop * 3 - start_frame
            if cds_first_stop < cds_starts[-1]:
                if ncorrected_frameshifts or nuncorrected_frameshifts:
                    E.warn( "nmd knockout transcript %s has frameshifts: %i corrected, %i uncorrected" %\
                                ( transcript_id,
                                  ncorrected_frameshifts, 
                                  nuncorrected_frameshifts) )
                is_nmd_knockout = True
                cds = peptide = ""
                lpeptide, lcds = 0, 0
                reference_first_stop_start, reference_first_stop_end = \
                    (map_cds2reference.mapRowToCol( cds_first_stop ),
                     map_cds2reference.mapRowToCol( cds_first_stop + 3 ) )
            elif pep_first_stop < len(peptide) - 1:
                is_stop_truncated = True
                cds = cds[:cds_first_stop]
                peptide[:pep_first_stop]
                lpeptide, lcds = len(peptide), len(cds)
                reference_first_stop_start, reference_first_stop_end = \
                    (map_cds2reference.mapRowToCol( cds_first_stop ),
                     map_cds2reference.mapRowToCol( cds_first_stop + 3 ) )
            else:
                E.warn( "first stop at %i(cds=%i) ignored: last exon start at %i" % \
                            (pep_first_stop, 
                             cds_first_stop,
                             cds_starts[-1]) )

        else:
            # -1 for no stop codon found
            pep_first_stop = -1
            cds_first_stop = -1
            lpeptide, lcds = len(peptide), len(cds)

        if peptide == None and nframeshifts == 0:
            E.warn( "transcript %s is knockout, though there are no indels - must be nonsense mutation" % (transcript_id))

        # build frames
        frames = [ start_frame ]
        start = start_frame
        l = 0 
        for end in cds_starts[1:]:
            l += end - start
            frames.append( (3 - l % 3) % 3 )
            start = end

        return Allele._make( (cds, 
                              peptide, 
                              len(cds_starts),
                              cds_starts,
                              exon_starts,
                              frames,
                              is_nmd_knockout,
                              is_splice_truncated,
                              is_stop_truncated,
                              nframeshifts,
                              ncorrected_frameshifts,
                              nuncorrected_frameshifts,
                              pep_first_stop,
                              lpeptide,
                              cds_first_stop,
                              lcds,
                              reference_first_stop_start,
                              reference_first_stop_end,
                              loriginal,
                              nsplice_noncanonical,
                              )), map_cds2reference

    if variant_exons or variant_introns:
        for allele in range(0,2):
            exons = dict( [(x,y[allele]) for x,y in variant_exons.iteritems() ] )
            introns = dict( [(x,y[allele]) for x,y in variant_introns.iteritems() ] )
            result.append( _buildAllele( allele, transcript, exons, introns, offsets[allele] ) )
    else:
        a = _buildAllele( 0, transcript, reference_exons, reference_introns, [] )
        result.append( a )
        result.append( a )

    return result



def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: gtf2alleles.py 2886 2010-04-07 08:47:46Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default]."  )
    parser.add_option("-t", "--tablename", dest="tablename", type="string",
                      help="tablename to get variants from (in samtools pileup format) [default=%default]."  )
    parser.add_option("-d", "--database", dest="database", type="string",
                      help="sqlite3 database [default=%default]."  )
    parser.add_option("-f", "--filename-exons", dest="filename_exons", type="string",
                      help="filename with transcript model information (gtf formatted file)  [default=%default]."  )
    parser.add_option("-r", "--filename-reference", dest="filename_reference", type="string",
                      help="filename with transcript models of a reference gene set. Stop codons that do not"
                      " overlap any of the exons in this file are ignore (gtf-formatted file)  [default=%default]."  )
    parser.add_option( "--filename-vcf", dest="filename_vcf", type="string",
                      help="filename with variants in VCF format. Should be indexed by tabix  [default=%default]."  )
    parser.add_option( "--filename-pileup", dest="filename_pileup", type="string",
                      help="filename with variants in samtools pileup format. Should be indexed by tabix  [default=%default]."  )
    parser.add_option( "--vcf-sample", dest="vcf_sample", type="string",
                      help="sample id for species of interest in vcf formatted file [default=%default]."  )
    parser.add_option("-s", "--filename-seleno", dest="filename_seleno", type="string",
                      help="filename of a list of transcript ids that are selenoproteins [default=%default]."  )
    parser.add_option("-m", "--module", dest="modules", type="choice", action="append",
                      choices=("gene-counts", "transcript-effects"),
                      help="modules to apply [default=%default]."  )
    parser.add_option("-o", "--output", dest="output", type="choice", action="append",
                      choices=("all", "peptide", "cds", "table", "gtf", "map"),
                      help="sections to output [default=%default]."  )
    parser.add_option("-k", "--with-knockouts", dest="with_knockouts", action="store_true",
                      help="add alleles that are knocked out to fasta and gtf files [default=%default]." )

    parser.set_defaults(
            genome_file = None,
            filename_exons = None,
            filename_referenec = None,
            filename_seleno = None,
            modules = [],
            border = 200,
            separator = "|",
            tablename = None,
            database = "csvdb",
            output = [],
            with_knockouts = False,
            filename_vcf = None,
            vcf_sample = None,
            )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv, add_output_options = True )

    ninput, nskipped, noutput = 0, 0, 0

    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta( options.genome_file )
    else:
        fasta = None

    if options.filename_seleno:
        seleno = set( IOTools.readList( open(options.filename_seleno, "r")) )
    else:
        seleno = {}

    infile_gtf = GTF.gene_iterator( GTF.iterator( options.stdin) )

    # acquire variants from SQLlite database
    if options.tablename:
        if not options.database:
            raise ValueError("please supply both database and tablename")
        variant_getter = VariantGetterSqlite( options.database, options.tablename )
    elif options.filename_pileup:
        variant_getter = VariantGetterPileup( options.filename_pileup )
    elif options.filename_vcf:
        variant_getter = VariantGetterVCF( options.filename_vcf, options.vcf_sample )
    else:
        raise ValueError("please specify a source of variants." )

    if len(options.output) == 0 or "all" in options.output:
        output_all = True
    else:
        output_all = False

    if "cds" in options.output or output_all:
        outfile_cds = E.openOutputFile( "cds.fasta" )
    else: outfile_cds = None

    if "map" in options.output or output_all:
        outfile_map = E.openOutputFile( "map.psl" )
    else: outfile_map = None

    if "peptide" in options.output or output_all:
        outfile_peptides = E.openOutputFile( "peptides.fasta" )
    else: outfile_peptides = None

    if "table" in options.output or output_all:
        outfile_alleles = E.openOutputFile( "table" )
        outfile_alleles.write( "\t".join( 
            ("gene_id", 
             "transcript_id", "allele_id", "contig", "strand", 
             "is_wildtype", 
             ("\t".join( Allele._fields)))) + "\n" )
    else: outfile_alleles = None

    if "gtf" in options.output or output_all:
        outfile_gtf = E.openOutputFile( "gtf" )
    else: outfile_gtf = None

    ## id separatar
    separator = options.separator

    for transcripts in infile_gtf:

        gene_id = transcripts[0][0].gene_id

        overall_start = min( [min( [x.start for x in y]) for y in transcripts] )
        overall_end = max( [max( [x.end for x in y]) for y in transcripts] )
        contig = transcripts[0][0].contig
        strand = transcripts[0][0].strand
        is_positive_strand = Genomics.IsPositiveStrand( strand )
        lcontig = fasta.getLength( contig )
        E.info( "%s: started processing on %s:%i..%i (%s)" % \
                    (gene_id, contig, overall_start, overall_end, strand ))

        ninput += 1
        extended_start = max( 0, overall_start - options.border )
        extended_end = min( lcontig, overall_end + options.border )

        # if contig.startswith("chr"): contig = contig[3:]

        variants = variant_getter( contig, extended_start, extended_end )

        E.debug( "%s: found %i variants in %s:%i..%i" % \
                     (gene_id, len(variants), contig, extended_start, extended_end ))

        if E.global_options.loglevel >= 10:
            print "# collected variants:", variants

        # collect intron/exon sequences
        # coordinates are forward/reverse
        # also updates the coordinates in transcripts
        all_exons, all_introns = collectExonIntronSequences( transcripts, fasta )

        # update variants such that they use the same coordinates
        # as the transcript
        variants = Variants.updateVariants( variants, lcontig, strand )

        # deal with overlapping but consistent variants
        variants = Variants.mergeVariants( variants )

        E.debug( "%s: found %i variants after merging in %s:%i..%i" % \
                     (gene_id, len(variants), contig, extended_start, extended_end ))

        if E.global_options.loglevel >= 10:
            print "# merged variants:", variants

        # collect coordinate offsets and remove conflicting variants
        variants, removed_variants, offsets = Variants.buildOffsets( variants, contig = contig ) 

        if len(removed_variants) > 0:
            E.warn("removed %i conflicting variants" % len(removed_variants) )
            for v in removed_variants:
                E.info("removed variant: %s" % str(v))

        E.info( "%i variants after filtering" % len(variants) )

        if len(variants) > 0:
            # build variants
            indexed_variants = Variants.indexVariants( variants )

            # update exon sequences according to variants
            variant_exons = buildVariantSequences( indexed_variants, all_exons )

            # update intron sequences according to variants
            variant_introns = buildVariantSequences( indexed_variants, all_introns )

            if E.global_options.loglevel >= 10:
                for key in variant_exons:
                    print "exon", key
                    Genomics.printPrettyAlignment( 
                        all_exons[key],
                        variant_exons[key][0],
                        variant_exons[key][1],
                        )
                for key in variant_introns:
                    print "intron", key
                    Genomics.printPrettyAlignment( 
                        all_introns[key][:30] + all_introns[key][-30:],
                        variant_introns[key][0][:30] + variant_introns[key][0][-30:],
                        variant_introns[key][1][:30] + variant_introns[key][1][-30:] )
        
        else:
            variant_exons, variant_introns = None, None


        for transcript in transcripts:

            transcript.sort( key = lambda x: x.start)

            transcript_id = transcript[0].transcript_id
            alleles = buildAlleles( transcript, 
                                    variant_exons, 
                                    variant_introns,
                                    all_exons,
                                    all_introns,
                                    offsets,
                                    is_seleno = transcript_id in seleno,
                                    reference_coordinates = False,
                                    )

            ##############################################################
            ##############################################################
            ##############################################################
            # output
            for aid, al in enumerate(alleles):

                allele, map_cds2reference = al

                reference_cds_sequence = buildCDSSequence( transcript, all_exons )
                is_wildtype = reference_cds_sequence == allele.cds

                allele_id = str(aid)
                assert len(allele.exon_starts) == allele.nexons
                assert len(allele.cds_starts) == allele.nexons
                assert len(allele.frames) == allele.nexons

                # the output id
                outid = separator.join( (gene_id, transcript_id, allele_id) )

                # output map between cds and reference
                if outfile_map and map_cds2reference:
                    match = Blat.Match()
                    match.mQueryId = allele_id
                    match.mQueryLength = allele.cds_len
                    match.mSbjctId = contig
                    match.mSbjctLength = lcontig
                    match.strand = strand
                    match.fromMap( map_cds2reference, use_strand = True )
                    outfile_map.write( "%s\n" % str(match) )
                    
                # only output sequences for genes that have not been knocked out, unless required
                if not allele.is_nmd_knockout or options.with_knockouts:

                    if outfile_gtf:
                        gtf = GTF.Entry()
                        gtf.gene_id = gene_id
                        gtf.transcript_id = transcript_id
                        gtf.addAttribute( "allele_id", allele_id)
                        gtf.contig = contig
                        gtf.strand = strand
                        gtf.feature = "CDS"
                        gtf.source = "gtfxnsps"
                        l = 0
                        last_cds_start = allele.cds_starts[0]
                        gtf.start = allele.exon_starts[0]
                        gtf.frame = allele.frames[0]

                        for exon_start, cds_start, frame in zip( allele.exon_starts[1:], 
                                                                 allele.cds_starts[1:], 
                                                                 allele.frames[1:]):
                            cds_length = cds_start - last_cds_start
                            gtf.end = gtf.start + cds_length
                            if not is_positive_strand:
                                gtf.start, gtf.end = lcontig - gtf.end, lcontig - gtf.start
                            outfile_gtf.write( str(gtf) + "\n" )

                            gtf.start = exon_start
                            gtf.frame = frame

                            l += cds_length
                            last_cds_start = cds_start

                        cds_length = len(allele.cds) - last_cds_start
                        gtf.end = gtf.start + cds_length
                        if not is_positive_strand:
                            gtf.start, gtf.end = lcontig - gtf.end, lcontig - gtf.start
                        outfile_gtf.write( str(gtf) + "\n" )

                    if outfile_cds:
                        outfile_cds.write(">%s\n%s\n" % (outid, allele.cds ))
                    if outfile_peptides:
                        outfile_peptides.write(">%s\n%s\n" % (outid, allele.peptide ))

                # reformat for tabular output
                allele = allele._replace( 
                    cds_starts = ",".join(map(str,allele.cds_starts)),
                    exon_starts = ",".join(map(str,allele.exon_starts)),
                    frames = ",".join(map(str,allele.frames)) )
                
                # convert reference coordinates to positive strand coordinates
                if allele.reference_first_stop_start >= 0 and not is_positive_strand:
                    allele = allele._replace(
                        reference_first_stop_start = lcontig - allele.reference_first_stop_end,
                        reference_first_stop_end = lcontig - allele.reference_first_stop_start, )
                    
                if outfile_alleles:
                    outfile_alleles.write( "%s\t%s\n" % (\
                            "\t".join( (gene_id, 
                                        transcript_id, 
                                        allele_id,
                                        contig,
                                        strand,
                                        "%i" % is_wildtype) ),
                            "\t".join(map(str,allele)) ) )
                            
                noutput += 1
                # only output first allele (debugging)
                # break

    E.info( "ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput,nskipped) )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )













####
####
##
##
## Copyright (C) 2008 Andreas Heger All rights reserved
##
## Author: Andreas Heger <heger@ebi.ac.uk>
##
## $Id: bed2bed.py 2861 2010-02-23 17:36:32Z andreas $
##
##
####
####
'''
wig2wig.py - manipulate wiggle files
====================================

Purpose
-------

manipulate wig-formatted files.

The script currently implements the following methods:

1. sanitize-genome: remove all empty intervals and intervals on unknown contigs. 
   Intervals extending beyond a contig a truncated. 
'''   

import sys
import re
import string
import optparse
import time
import os
import itertools
import tempfile
import subprocess
import shutil

import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.IOTools as IOTools

def sanitizeGenome( infile, outfile, contigs ):
    """truncate bed intervals that extend beyond contigs.

    removes empty intervals (start == end).

    throws an error if start > end.
    """
    
    ninput, noutput = 0, 0
    ntruncated_contig, nskipped_contig, nskipped_empty = 0, 0, 0

    for line in infile:
        if line.startswith( "track"):
            outfile.write(line)
            continue
        elif line.startswith( "variableStep" ):
            contig = re.search( "chrom=(\S+)", line ).groups()[0]
            span = int(re.search( "span=(\d+)", line ).groups()[0])
            
            if contig not in contigs:
                nskipped_contig += 1
                skip_contig = True
            else:
                skip_contig = False
                outfile.write(line)
                contig_size = contigs[contig]

            continue

        if skip_contig: continue
    
        start, val = map(int, line[:-1].split("\t") )
        
        if start + span > contig_size:
            ntruncated_contig += 1
            continue
        
        outfile.write( line )

    E.info( "ninput=%i, noutput=%i, nskipped_contig=%i, ntruncated=%i, nskipped_empty=%i" % \
                (ninput, noutput, nskipped_contig, ntruncated_contig, nskipped_empty) )


def main( argv = sys.argv ):

    parser = E.OptionParser( version = "%prog version: $Id: bed2bed.py 2861 2010-02-23 17:36:32Z andreas $", 
                                    usage = globals()["__doc__"] )


    parser.add_option( "-m", "--method", dest="methods", type="choice", action="append",
                       choices=("merge", "filter-genome", "bins", "block", "sanitize-genome", "shift", "extend" ),
                       help="method to apply [default=%default]"  )

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome."  )

    parser.set_defaults( methods = [],
                         merge_distance = 0,
                         binning_method = "equal-bases",
                         genome_file = None,
                         bam_file = None,
                         num_bins = 5,
                         merge_min_intervals = 1,
                         bin_edges = None,
                         offset = 10000,
                         test = None,
                         extend_distance=1000)
    
    (options, args) = E.Start( parser, add_pipe_options = True )

    contigs = None

    if options.genome_file:
        genome_fasta = IndexedFasta.IndexedFasta( options.genome_file )
        contigs = genome_fasta.getContigSizes()

    for method in options.methods:
        if method ==  "filter-genome":
            if not contigs: raise ValueError("please supply contig sizes" )
            processor = filterGenome( processor, contigs )
        elif method == "sanitize-genome":
            if not contigs: raise ValueError("please supply contig sizes" )
            processor = sanitizeGenome( options.stdin, options.stdout, contigs )

    E.Stop()


if __name__ == "__main__":
    sys.exit( main( sys.argv ) )
